/* Krister Swenson                                                  Spring 2006

Generate all sorting inversions and output the sorting graph in graphviz format.
  `-> run "allrevs > graph.dot; dot -Tps graph.dot -o graph.ps"

Also check whether all intermediate permutations are the same distance from
the identity. ---> don't think it does this right now :) .

Keep track of the sequence "renamed" and output it's distance to the identity
as part of the node name as well.
*/

#include <iostream>
#include <vector>
#include "../graph/med.h"
#include "../seqtools/seqfuncs.h"

using namespace std; 

#define PRINT_BOTH 0     //If 1 both strings are printed.
#define PRINT_RENAMED 0  //If 1 the renamed string is printed instead of targ.
#define PRINT_THIRD 0    //If 1 the target is printed according to the third.
#define PRINT_DISTS 0    //If 1 the distances are put under the permutation(s).
#define PRINT_UNSAFE 0   //Also print the dead-end unsafe nodes.
#define COLOR_DE 1       //Color the dead-end nodes.
#define COLOR_CS 0       //Color the cyclesplitting edges.
#define COLOR_OR 0       //Color the oriented edges.
#define LANDSCAPE 0      //Print in landscape mode if not 0.
float RANKSEP = 0;       //The seperation between the lines. Automatic if 0.

#if PRINT_BOTH && PRINT_THIRD
  char renstr[] = "BOTH_THIRD";
#elif PRINT_BOTH
  char renstr[] = "BOTH";
#elif PRINT_RENAMED
  char renstr[] = "RENAMED";
#elif PRINT_THIRD
  char renstr[] = "TARGET_THIRD";
#else
  char renstr[] = "TARGET";
#endif

void printSeq(int s[], int slen);
void printVector(const vector<Pair> &v);
void allRevs(Ograph &gr, const char* str, const Ograph &rgr);
void allRevs(Ograph &gr, const Ograph &rgr,
             const int third[], const int thirdlen);
char* getGraphString(const Ograph &g, const int third[], const int thirdlen);
void printSeqString(const char* str, const char* renamedstr);
int* renameSeq(int s1[], int s1len, int s2[], int s2len);





int main()
  {
  //int targ[] = {-4,5,1,3,2,6};         //(a)
  //int renamed[] = {-4,5,1,3,2,6};
  //int targ[] = {-4,-6,-2,-3,-1,-5};    //(b)Out one from a.
  //int renamed[] = {-4,-6,-2,-3,-1,-5};
  //int targ[] = {2,6,4,-3,-1,-5};    //Down  one from b.
  //int renamed[] = {2,6,4,-3,-1,-5};

  //int targ[] = {-4,-3,-1,-5,2};        //sublattice of a with separation.
  //int renamed[] = {-4,-3,-1,-5,2};
  //int targ[] = {-3,5,-2,-1,-4};        //separation
  //int renamed[] = {-3,5,-2,-1,-4};
  //int targ[] = {-3,5,-1,4,2};          //separation
  //int renamed[] = {-3,5,-1,4,2};


  //int targ[] = {2,1,-7,-8,3,6,9,-4,-5,11,10};
  //int renamed[] = {2,1,-7,-8,3,6,9,-4,-5,11,10};
  //int targ[] = {-6,-7,1,-8,-5,-2,9,-3,-4};
  //int renamed[] = {-6,-7,1,-8,-5,-2,9,-3,-4};

//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
//:noninterfering:
  //int targ[] = {-2,3,1};   //seq1
  //int renamed[] = {-2,3,1};
  //int third[] = {-2,3,1};
  //int targ[] = {2,-1,-3};   //seq2
  //int renamed[] = {2,-1,-3};
  //int third[] = {2,-1,-3};
  

//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
//:other:
  //int targ[] = {-2,1,3,4,6,-5};
  //int renamed[] = {-2,1,3,4,5,6};
  //int third[] = {1,2,3,4,6,-5};
  //int targ[] = {-3,2,1,4,-7,6,5};
  //int renamed[] = {-3,2,1,4,-7,6,5};
  //int third[] = {-3,2,1,4,-7,6,5};

  //int targ[] = {2,-3,5,1,4};
  //int renamed[] = {2,-3,5,1,4};
  //int third[] = {2,-3,5,1,4};
  //int targ[] = {4,1,-2,5,3};
  //int renamed[] = {4,1,-2,5,3};
  //int third[] = {4,1,-2,5,3};
  //int targ[] = {2,-5,3,-1,4};
  //int renamed[] = {2,-5,3,-1,4};
  //int third[] = {2,-5,3,-1,4};

//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
//:largest signature:
  //int targ[] = {3,-1,2,4};     //1
  //int renamed[] = {3,-1,2,4};  //1
  //int third[] = {3,-1,2,4};  //1

  //int targ[] = {-3,-2,4,1};    //2
  //int renamed[] = {-3,-2,4,1}; //2
  //int third[] = {-3,-2,4,1}; //2
  //int targ[] = {-1,-3,4,-2};   //2 according to 1
  //int renamed[] = {-3,-2,4,1}; //2


  //int targ[] = {2,-5,-4,-3,1,6,7};     //dead-end
  //int renamed[] = {2,-5,-4,-3,1,6,7};
  //int targ[] = {-2,-5,-4,-6,-1,3,7};     //dead-end2
  //int renamed[] = {-2,-5,-4,-6,-1,3,7};

  //int targ[] = {4,-2,6,-3,1,-5,7};  //dead-end3
  //int renamed[] = {4,-2,6,-3,1,-5,7};
  //int targ[] = {-2,6,-4,-3,-7,-1,-5};  //dead-end4
  //int renamed[] = {-2,6,-4,-3,-7,-1,-5};
//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
//:inverse:
  //int targ[] = {-1,-6,2,-5,-4,-3};    //2
  //int renamed[] = {-1,3,-6,-5,-4,-2}; //2.inv

  //int targ[] = {-3,2,4,1};            //3
  //int renamed[] = {4,2,-1,3};         //3
  //int targ[] = {2,-4,3,-1};           //1
  //int renamed[] = {-4,1,3,-2};        //1.inv
  //int targ[] = {-4,1,3,-2};           //1I
  //int renamed[] = {2,-4,3,-1};        //1I.inv
//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
//:benign:
  //int targ[] = {3,-1,2};       //(h)
  //int renamed[] = {3,-1,2};    //Two that are one (benign) inv away and edit 
  //int targ[] = {-3,-1,2};      // paths don't cross till the ID.
  //int renamed[] = {-3,-1,2};   //(7)
  //int targ[] = {-3,-1,-2};     //(1a)
  //int renamed[] = {-3,-1,-2};  //(1a)
  //int targ[] = {-3,2,1};       //(a1)
  //int renamed[] = {-3,2,1};    //(a1)

  //int targ[] = {-3,-1,2};      //(7)
  //int targ[] = {-3,2,1};       //(a1)
  //int targ[] = {3,-1,2};       //(h)
  //int acc[] = {-3,-1,2};       //Make renamed according to (7).
  //int acc[] = {-3,2,1};        //Make renamed according to (a1).
  //int third[] = {-3,2,1};      //(a1)
  //int third[] = {3,-1,2};      //(h)
                                 //`-> so the targ will be replaced by the 
                                 //sequence named according to (h).


                                 //Three on an edit path from (h) to the id:
  //int targ[] = {-1,3,-2};      //(d) labeled according to (5)
  //int renamed[] = {-3,-2,1};   //(d)
  //int targ[] = {-1,-2,-3};     //(e) labeled according to (5)
  //int renamed[] = {-3,1,2};    //(e)
  //int targ[] = {3,-2,1};       //(g) labeled according to (5)
  //int renamed[] = {-2,1,3};    //(g)

//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
  //int targ[] = {-1,-2,-3,-4,-5};           //Caterpillar
  //int renamed[] = {-1,-2,-3,-4,-5};
  //int targ[] = {-1,3,2,-4,-5};             //Caterpillar Benign
  //int renamed[] = {-1,3,2,-4,-5};
  //int targ[] = {4,3,2,1,-5};               //Caterpillar Benign2
  //int renamed[] = {4,3,2,1,-5};
  //int targ[] = {-5,7,-3,9,-1,6,-4,8,-2};   //Single Catalyst
  //int renamed[] = {-5,7,-3,9,-1,6,-4,8,-2};
  //int targ[] = {-5,7,-3,9,2,-8,4,-6,1};    //Single Catalyst Benign
  //int renamed[] = {-5,7,-3,9,2,-8,4,-6,1};
  //int targ[] = {-3,5,-1,4,-2};             //Single Catalyst (too small)
  //int renamed[] = {-3,5,-1,4,-2};
  //int targ[] = {-4,-6,3,7,2,-5,1};         //Tannier
  //int renamed[] = {-4,-6,3,7,2,-5,1};
  //int targ[] = {-4,-6,-2,-7,-3,-5,1};      //Tannier Benign
  //int renamed[] = {-4,-6,-2,-7,-3,-5,1};
  //int targ[] = {5,-2,-7,-3,6,4,1};         //Tannier Benign2
  //int renamed[] = {5,-2,-7,-3,6,4,1};
  //int targ[] = {2,4,-1,-3,5};              //little Shamir
  //int renamed[] = {2,4,-1,-3,5};
  //int targ[] = {2,4,6,8,-1,-3,-5,-7,9};    //Shamir
  //int renamed[] = {2,4,6,8,-1,-3,-5,-7,9};
  //int targ[] = {2,-6,-4,8,-1,-3,-5,-7,9};  //Shamir Benign
  //int renamed[] = {2,-6,-4,8,-1,-3,-5,-7,9};
  //int targ[] = {2,4,6,8,-1,7,5,3,9};       //Shamir B3
  //int renamed[] = {2,4,6,8,-1,7,5,3,9};
  //int targ[] = {-4,1,-5,2,-6,3};           //shamir.b.renamed.6
  //int renamed[] = {-4,1,-5,2,-6,3};
  //int targ[] = {-4,1,6,2,-5,3};            //shamir.b.renamed.6
  //int renamed[] = {-4,1,6,2,-5,3};




  //int targ[] = {-5,1,-4,-3,-2};         //first
  //int renamed[] = {2,-1,-5,-4,-3};      //first (1 OPath, 5 ConcaveP)
  //int targ[] = {1,-6,2,-5,-4,-3,7};     //first1/2
  //int renamed[] = {1,3,2,5,4,7,6};      //first1/2
  //int targ[] = {4,-1,5,-3,-2};          //second
  //int renamed[] = {1,2,3,4,5,6};        //----------DUMMY---------------------
  //int targ[] = {-6,1,2,-4,-3,-5};       //third
  //int targ[] = {-6,1,3,2,4,-5};         //hurdle

  //int targ[] = {1,2,-9,-8,-6,-5,7,3,4,-12,-11,-10,13,-16,-15,-14,17,18,19};//0
  //int acc[] = {1,2,-9,-8,-6,-5,7,3,4,-12,-11,-10,13,-16,-15,-14,17,18,19};
  //int renamed[] = {1,2,-9,-8,-6,-5,7,3,4,-12,-11,-10,13,-16,-15,-14,17,18,19};
  //int targ[] = {1,2,3,-5,-4,6,8,9,-7,-12,-11,-10,13,14,15,16,17,18,19};//1
  //int third[] = {1,2,3,-5,-4,6,8,9,-7,-12,-11,-10,13,14,15,16,17,18,19};
  //int renamed[] = {1,2,3,-5,-4,6,8,9,-7,-12,-11,-10,13,14,15,16,17,18,19};
  //int targ[] = {1,2,3,-11,-4,7,-12,-5,-10,-6,-19,-17,-15,-8,13,14,9,18,-16};//2
  //int renamed[] = {1,2,3,-11,-4,7,-12,-5,-10,-6,-19,-17,-15,-8,13,14,9,18,-16};
  //int targ[] = {1,2,-10,-6,4,11,-3,5,12,-7,-9,-8,13,14,-18,-17,19,16,15};//3
  //int third[] = {1,2,-10,-6,4,11,-3,5,12,-7,-9,-8,13,14,-18,-17,19,16,15};
  //int renamed[] = {1,2,-10,-6,4,11,-3,5,12,-7,-9,-8,13,14,-18,-17,19,16,15};
  //int targ[] = {5,6,7,-1,-4,-3,10,11,-12,-2,8,9,13,-15,14,16,17,19,18};//4
  //int third[] = {5,6,7,-1,-4,-3,10,11,-12,-2,8,9,13,-15,14,16,17,19,18};
  //int renamed[] = {5,6,7,-1,-4,-3,10,11,-12,-2,8,9,13,-15,14,16,17,19,18};
  //int targ[] = {5,6,7,-1,-4,-3,-2,-11,-10,-9,-8,12,-18,-19,-17,-16,-15,-14,-13};//5
  //int renamed[] = {5,6,7,-1,-4,-3,-2,-11,-10,-9,-8,12,-18,-19,-17,-16,-15,-14,-13};

  //int targ[] = {1,2,-9,-8,-6,-5,7,3,4,-12,-11,-10,13};    //0short
  //int renamed[] = {1,2,-9,-8,-6,-5,7,3,4,-12,-11,-10,13}; //0short
  //int targ[] = {1,2,3,-5,-4,6,8,9,-7,-12,-11,-10,13};     //1short
  //int renamed[] = {1,2,3,-5,-4,6,8,9,-7,-12,-11,-10,13};  //1short
  //int targ[] = {1,2,3,4,5,6,8,9,-7,-12,-11,-10,13};       //0.1shortCenter
  //int renamed[] = {1,2,3,4,5,6,8,9,-7,-12,-11,-10,13};    //0.1shortCenter

  //int targ[] = {5,6,7,-1,-4,-3,10,11,-12,-2,8,9,13};        //4short
  //int renamed[] = {5,6,7,-1,-4,-3,10,11,-12,-2,8,9,13};     //4short
  //int targ[] = {5,6,7,-1,-4,-3,-2,-11,-10,-9,-8,12,13};        //5short
  //int renamed[] = {5,6,7,-1,-4,-3,-2,-11,-10,-9,-8,12,13};     //5short
//-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o


  //int targlen = sizeof(targ)/sizeof(int);       //The sequence length.
  //int renamedlen = sizeof(renamed)/sizeof(int); //The renamed sequence length.

  int *targ=0;
  int *renamed=0;
  
  cerr << "TARGET:\t";
  int targlen = getInputSeq(&targ);

  // make renamed according to acc:
  //int acclen = sizeof(acc)/sizeof(int);         //The acc sequence length.
  //int* renamed = renameSeq(targ,targlen,acc,acclen); int renamedlen = targlen;

#if PRINT_BOTH || PRINT_RENAMED
  cerr << "RENAMED:";
  int renamedlen = getInputSeq(&renamed);
#else
  int renamedlen = targlen;
  renamed = copySeq(targ, targlen);
#endif
#if PRINT_BOTH
  int *third=0;
  cerr << "THIRD:\t";
  int thirdlen = getInputSeq(&third);
  //int thirdlen = sizeof(third)/sizeof(int);     //The third sequence length.
#endif


  cerr << "TARGET:\t";
  printSeq(targ, targlen);
  cerr << "\nRENAMED:";
  printSeq(renamed, renamedlen);
  cerr << endl;

  Ograph gr(targ, targlen);       //Build the graph.
  Ograph rgr(renamed, renamedlen);//Build the graph.

  /*
  char* inverse = gr.getRenStr(third, thirdlen);
  cerr << "\nTaccTH:\t" << inverse << endl;
  free(inverse);
  */

  if(!RANKSEP)
    {
    int dist = gr.getDist();
    RANKSEP = 1.2*dist*dist - 2.4*dist - 7;
    }

  cout << "digraph " << renstr
       << "\n  {\n  "
       << "size = \"10,7\";";
  if(LANDSCAPE)
    cout<< "orientation=landscape;";
  else
    cout<< "orientation=portrait;";
  cout << "center=false;"
       << "concentrate=true;"
       << "ranksep=" << RANKSEP << ";"
       << "bgcolor=transparent;"
       << endl;
#if PRINT_BOTH
  allRevs(gr, rgr, third, thirdlen);
#else
     //   DO THIS TO MAKE IT FASTER BUT PRINT_BOTH won't work:
  char *seqstr;
  if(PRINT_RENAMED)
    seqstr = rgr.getSeqStr();
  else
    seqstr = gr.getSeqStr();
  allRevs(gr, seqstr, rgr);
  free(seqstr);
#endif
  cout << "  }" << endl;

  free(targ);
  free(renamed);
#if PRINT_BOTH
  free(third);
#endif
  return(0);
  }


/* Print every link of the sorting tree in graphviz format.
*/
void allRevs(Ograph &gr, const Ograph &rgr,
             const int third[], const int thirdlen)
  {
  vector<Pair> oOps;

  if(gr.notSorted())                   //If the sequence is not sorted,
    {
    gr.getSafeInvs(oOps);              //get the safe ops on divergent edges,

    char *grstr = getGraphString(gr, third, thirdlen);
    char *rgrstr = rgr.getSeqStr();

    for(int i=0; size_t(i)<oOps.size(); i++)
      {                                //for each divergent op
      Ograph newgr = gr;               //make a new graph
      newgr.revAll(oOps[i]);           //that we do one of the inversions on.
      Ograph newrgr = rgr;             //make a new renamed graph
      newrgr.revAll(oOps[i]);          //that we do one of the inversions on.

      cout << "  \"";
      printSeqString(grstr, rgrstr);
      if(PRINT_DISTS)
        cout << "\\n" << gr.getDistNH() << "(" << rgr.getDistNH() << ")";
      if(PRINT_THIRD)
        {
        int* tmp = gr.getRenSeq(third, thirdlen);
        Ograph tmpgr(tmp, gr.getSeqLen(), true);
        if(PRINT_DISTS)
          cout << tmpgr.getDistNH();
        free(tmp);
        }
      cout << "\" -> \"";

      char *newgrstr = getGraphString(newgr, third, thirdlen);
      char *newrgrstr = newrgr.getSeqStr();
      printSeqString(newgrstr, newrgrstr);
      free(newgrstr);
      free(newrgrstr);
      if(PRINT_DISTS)
        cout << "\\n" << newgr.getDistNH() << "(" << newrgr.getDistNH() << ")";
      if(PRINT_THIRD)
        {
        int* tmp = newgr.getRenSeq(third, thirdlen);
        Ograph tmpgr(tmp, newgr.getSeqLen(), true);
        if(PRINT_DISTS)
          cout << tmpgr.getDistNH();
        free(tmp);
        }
      cout << "\";\n";


      allRevs(newgr, newrgr, third, thirdlen);   //and continue to sort.
      }
    free(grstr);
    free(rgrstr);
    }
  }


/* Print the appropriate combination of the given strings to cout.
*/
void printSeqString(const char* str, const char* renamedstr)
  {
  if(PRINT_BOTH)
    cout << str << "\\n" << renamedstr;
  else if(PRINT_RENAMED)
    cout << renamedstr;
  else
    cout << str;
  }

/* Return the string for the given graph (really just to hide the PRINT_THIRD).
*/
char* getGraphString(const Ograph &g, const int third[], const int thirdlen)
  {
  if(PRINT_THIRD)
    return g.getRenStr(third, thirdlen);
  else
    return g.getSeqStr();
  }

//--------THIS IS THE FASTER VERSION THAT DOESN'T ALLOCATE THE STRING A LOT-----
/* Print every link of the sorting tree in graphviz format.
 ** PRINT_BOTH won't work with this version.
*/
void allRevs(Ograph &gr, const char* str, const Ograph &rgr)
  {
  vector<Pair> oOps;

  if(gr.notSorted())                   //If the sequence is not sorted,
    {
    if(PRINT_UNSAFE)
      gr.getCSInvs(oOps);              //get the ops on divergent edges,
    else
      gr.getSafeInvs(oOps);            //get the safe ops on divergent edges,
    if(COLOR_DE && oOps.empty())
      cout << "  \"" << str << "\\n\"[color=blue,style=filled];\n";
    for(int i=0; size_t(i)<oOps.size(); i++)
      {                                //for each divergent op
      Ograph newgr = gr;               //make a new graph
      newgr.revAll(oOps[i]);           //that we do one of the inversions on.
      Ograph newrgr = rgr;             //make a new graph
      newrgr.revAll(oOps[i]);          //that we do one of the inversions on.
      char *newstr;
      if(PRINT_RENAMED)
        newstr = newrgr.getSeqStr();   //Output the sequence
      else
        newstr = newgr.getSeqStr();    //Output the sequence
      if(PRINT_THIRD)
        {
        cout << "  \"" << str << "\\n" << gr.getDistNH() << "("
             << rgr.getDistNH() << ")\" -> \""
             << newstr << "\\n" << newgr.getDistNH() << "("
             << newrgr.getDistNH() << ")\";\n";
        }
      else
        {
        if(gr.createsAdjacency(oOps[i])) //If this is an oriented edge
          {
          if(COLOR_OR)                   //and we should print in new color:
            cout << "  \"" << str << "\\n\" -> \"" << newstr << "\\n\""
                 << "[color=yellow,style=bold];\n";
          else
            cout << "  \"" << str << "\\n\" -> \"" << newstr << "\\n\";\n";
          }
        else if(COLOR_CS)                //Otherwise print the divergent op:
          cout << "  \"" << str << "\\n\" -> \"" << newstr << "\\n\""
               << "[color=red,style=bold];\n";
        else
          cout << "  \"" << str << "\\n\" -> \"" << newstr << "\\n\";\n";
        }

      allRevs(newgr, newstr, newrgr);  //and continue to sort.
      free(newstr);
      }
    }
  }



/* Print a sequence.
*/
void printSeq(int s[], int slen)
  {
  for(int i=0; i<slen; i++)
    cerr << s[i] << " ";
  }

/* Print a vector.
*/
void printVector(const vector<Pair> &v)
  {
  for(int i=0; size_t(i)<v.size(); i++)
    cerr << v[i];
  cerr << endl;
  }

/* Return s1 renamed as though s2 were the target.
*/
int* renameSeq(int s1[], int s1len, int s2[], int s2len)
  {
  int* ret = (int*)calloc(s1len, sizeof(int));

  assert(s1len == s2len);    //Ensure the sequence lengths match.

  int index[s2len];          //index[i] is the index where i is located in s2.
                             // (so should be the value that i becomes in s1)
  for(int i=0; i<s2len; i++)     //Build the index:
    {
    if(s2[i] < 0)                //If the number is negative
      index[abs(s2[i])] = -(i+1);//then store its index as negative
    else                         //otherwise,
      index[s2[i]] = (i+1);      //store it as positive.
    }

  for(int j=0; j<s1len; j++)     //Use the index.
    {
    if(s1[j]>0)
      ret[j] = index[abs(s1[j])];
    else
      ret[j] = -index[abs(s1[j])];
    }

  return ret;
  }
