/* Krister Swenson                                                  Fall 2008

Draw the sorting lattice including all permutations D inversions from the
identity.
  ** doesn't create hurdles as it should.
*/

#include <stdlib.h>
#include <iostream>
#include <vector>
#include <string>
#include "../graph/Ograph.h"
#include "../seqtools/seqfuncs.h"

using namespace std; 

#define N 4     //The length of the permutation.
#define D 3     //The distance from the identity every permutation is within.

#define RANKSEP 1  //Distance between nodes.

void printUnsortingInvs(Ograph &p);
void printEdge(vector<int> p, vector<int> q);
//void printSeq(int s[], int slen);
//void printVector(const vector<ipair> &v);

template < class T >
string ToString(const T &arg)
  {
  ostringstream out;
  out << arg;
  return(out.str());
  }


int main()
  {
  Ograph p(makeIDv(N));

  string graphname = ToString(N) + "." + ToString(D);
  cout << "graph " << graphname
       << "\n  {\n  "
       << "size = \"10,7\";"
       << "orientation=portrait;"
       << "root = \"" << p.getSeqVec() << "\";"
       << "concentrate=false;"
       // << "overlap=false;"
       << "overlap=scale;"
       // << "splines=true;"
       // << "ranksep=" << RANKSEP << ";"
       << "bgcolor=transparent;"
       << endl;

  printUnsortingInvs(p);

  cout << "  }\n";
  return 0;
  }



/*Print all permutations that are 1 unsorting inverion from p.
*/
void printUnsortingInvs(Ograph &p)
  {
  if(p.getDistNH() >= D)
    return;

  vector<int> pv = p.getSeqVec();
  vector<Pair> usr; 
  p.getUnsortingInvs(usr);
  for(vector<Pair>::iterator i=usr.begin(); i<usr.end(); i++)
    {
    p.revAll(*i);
    printEdge(pv, p.getSeqVec());
    printUnsortingInvs(p);
    p.revAll(*i);
    }
  }


/* Print the edge from one permutation to the other.
*/
void printEdge(vector<int> p, vector<int> q)
  { cout << "  \"" << p << "\" -- \"" << q << "\";\n"; }

/* Print the arc (directed) from one permutation to the other.
*/
void printArc(vector<int> p, vector<int> q)
  { cout << "  \"" << p << "\" -> \"" << q << "\";\n"; }



/* Print a sequence.
*/
void printSeq(int s[], int slen)
  {
  for(int i=0; i<slen; i++)
    cerr << s[i] << " ";
  }

/* Print a vector.
*/
//void printVector(const vector<ipair> &v)
//  {
//  for(int i=0; size_t(i)<v.size(); i++)
//    cerr << v[i];
//  cerr << endl;
//  }
