// Vaibhav Rajan and Krister Swenson                              Summer 2008
//
//Maintains (through reversals) a permutation in a splay tree.
//Each node holds the maximum and minimum postitive and negative elements in
//its subtree so that an oriented inversion can be quickly found.
//
//Reversal flags are kept at the nodes so that reversals take only lg(n) time.
//At any time the invariant is kept that a node with the -reversed- flag to
//true will have its left subtree on it's right (and vice-versa), its
//minPositive as its minNegative (and vice-versa), its element reversed,
//and its (and it's subtrees') index(ices) out of date.
//


#include "Permutation.h"


using namespace std;

/*Innitialize to the ID.
*/
Permutation::Permutation(int len)
  { nullNode = NULL; initWith(makeIDv(len)); }

Permutation::Permutation(int* seq, int slen)
  { nullNode = NULL; initWith(seq, slen); }


Permutation::Permutation(vector<int> seq)
  {
  int* s = new int[seq.size()];
  for(unsigned int i=0; i < seq.size(); i++)
    s[i] = seq[i];

  nullNode = NULL;
  initWith(s, seq.size());
  delete [] s;
  }

void Permutation::initWith(int* seq, int slen)
{
    if(nullNode != NULL)
      makeEmpty();
         //Setup the data structure:
    nullNode = new BinaryNode;
    nullNode->left = nullNode->right = nullNode->parent = nullNode;
    root = nullNode;

    lookup.resize(slen+1);

    mem = new BinaryNode[slen];

    int j = 0;
         //Add the elements of the permutation:
    for(int i=0; i<slen; i++)
      insert(i,seq[i],j);
}

void Permutation::initWith(vector<int> seq)
{
    if(nullNode != NULL)
      makeEmpty();
         //Setup the data structure:
    nullNode = new BinaryNode;
    nullNode->left = nullNode->right = nullNode->parent = nullNode;
    root = nullNode;

    lookup.resize(seq.size()+1);

    mem = new BinaryNode[seq.size()];

    int j = 0;
         //Add the elements of the permutation:
    for(unsigned i=0; i<seq.size(); i++)
      insert(i,seq[i],j);
}





/*
void Permutation::initWith(int* seq, int slen)
{
    if(nullNode != NULL)
        makeEmpty();
         //Setup the data structure:
    nullNode = new BinaryNode;
    nullNode->left = nullNode->right = nullNode->parent = nullNode;
    root = nullNode;

    lookup.resize(slen+1);

         //Add the elements of the permutation:
    for(int i=0; i<slen; i++)
      insert(i,seq[i]);
}

void Permutation::initWith(vector<int> seq)
{
    if(nullNode != NULL)
        makeEmpty();
         //Setup the data structure:
    nullNode = new BinaryNode;
    nullNode->left = nullNode->right = nullNode->parent = nullNode;
    root = nullNode;

    lookup.resize(seq.size()+1);

         //Add the elements of the permutation:
    for(unsigned i=0; i<seq.size(); i++)
      insert(i,seq[i]);
}
*/

/*
Permutation::Permutation( const Permutation & rhs )
{
    nullNode = new BinaryNode;
    nullNode->left = nullNode->right = nullNode->parent = nullNode;
    root = nullNode;
    *this = rhs;
}
*/

Permutation::~Permutation( )
{
    makeEmpty( );
}

/* Undo reversals until the most recent unsafe.
*/
void Permutation::fixUnsafe(int &k, vector<int> &original, Permutation &op,
                            list<ipair> &beginning, list<ipair> &ending)
  {
    list<ipair> R = getUnstuck(op,beginning);//Backtrack if needed
                                             // (store the revs in R).

    if(!R.empty())
      {                                      //If there was an unsafe reversal
      k++;
      ipair r = R.front();                   //then get it.
      R.pop_front();

      vector<int> p = getSeqVec();
      removeOrientedComps(p);                  //Shuffle the elements so that
      initWith(p);                             //we sort to the new target.
      ending.splice(ending.begin(), R);        //Put R on the front of ending.

      ipair brokencomp = getBrokenComp(r);     //The comp broken by unsafe rev.
      rev(r);                                  //Undo the unsafe reversal.

      ipair newrev = getSafeRev(r, brokencomp);
      rev(newrev);                             //Now do a safe reversal instead.
      op.rev(newrev);
      beginning.push_back(newrev);
      }

  }

/* Finish filling out the reversal indicies from the list of oriented pairs.
*/
void Permutation::finalizeSortSeq(list<ipair> &sortseq, Permutation &p,
                                  list<ipair> &beginning, list<ipair> &ending)
  {
  for(list<ipair>::iterator i=beginning.begin(); i!=beginning.end(); i++)
    {
    sortseq.push_back(*i);
    p.rev(*i);                                //Do the beginning reversals.
    }
  for(list<ipair>::iterator i=ending.begin(); i!=ending.end(); i++)
    {
    BinaryNode *left=0, *right=0;
    if(i->first != 0)
      {
      left = p.lookup[i->first];
      p.updateNodeInfo(left);                 //Make sure the index is true.
      }
    if(i->second != p.length()+1)
      right = p.lookup[i->second];

    if(i->first != 0 && left->element < 0)    //If the smaller one is negative
      *i = p.getOtherEndpoint(left, USE_MAX);
    else                                      //save the reversal to be done.
      {                                       //Otherwise larger is negative:
      p.updateNodeInfo(right);                //Make sure the index is true.
      *i = p.getOtherEndpoint(right, USE_MIN);
      }                                       //Save the reversal to be done.

    p.rev(*i);                                //Now apply the reversal.
    sortseq.push_back(*i);
    }
  }


/* Return a sorting sequence.
*/
list<ipair> Permutation::getSortSeq(int flag)
  {
  int k=0;
  vector<int> original = getSeqVec();
  Permutation op(original);                  //Keep the original permutation.
  list<ipair> beginning;
  list<ipair> ending;

  while(!isSorted())
    {
    ipair inv;
    while(nextRev(inv, flag))
      {
      beginning.push_back(inv);
      op.rev(inv);
      }

    fixUnsafe(k, original, op, beginning, ending);
    }

  if(ending.empty())             //If k=0 then just return:
    {
    cout << "k = " << k << endl;
    return beginning;
    }

    //Now that we have the beginning and ending of our sort sequence we will
    // turn the ending (which are adjacencies created) into reversal indices: 
  list<ipair> sortseq;
  Permutation p(original);
  finalizeSortSeq(sortseq, p, beginning, ending);

  cout << "k = " << k << endl;
  return sortseq;
  }



/* Find the HP graph and turn all oriented components into adjacencies.
   Return the corresponding permutation.
*/
struct Node
  {
  int reality;  //The reality edge.
  int desire;   //The desire edge.
  int label;    //The node label.
  };
void Permutation::removeOrientedComps(vector<int> &p) const
  {
  vector<int> seq;
  seq.push_back(0);
  for(vector<int>::iterator i=p.begin(); i!=p.end(); i++)
    seq.push_back(*i);
  seq.push_back(p.size()+1);

  vector<Node> gr;
  int graphsize = 2*(int)p.size() + 2;
  gr.resize(graphsize);

  vector<int>
    nodeLoc(graphsize);       //Location of node i in the graph:0,-1,1,-2,2...

  nodeLoc[0] = 0;             //Initialize the frame(two added nodes):
  nodeLoc[graphsize-1] = graphsize-1;

  gr[0].label = 0;
  gr[0].reality = 1;
  for(unsigned i=1; i <= p.size(); i++)
    {
    if(seq[i] > 0)
      {
      nodeLoc[abs(seq[i]*2)] = 2*i;
      nodeLoc[abs(seq[i]*2)-1] = (2*i)-1;
      }
    else
      {
      nodeLoc[abs(seq[i]*2)-1] = 2*i;
      nodeLoc[abs(seq[i]*2)] = (2*i)-1;
      }

    if(seq[i] < 0)                         //THIS IS Bergeron CONVENTION:
      {
      gr[(2*i)-1].label = -2*seq[i];
      gr[2*i].label = (-2*seq[i])-1;
      }
    else
      {
      gr[(2*i)-1].label = (2*seq[i])-1;
      gr[2*i].label = 2*seq[i];
      }

    gr[(2*i)-1].reality = 2*(i-1);
    gr[2*i].reality = (2*(i+1))-1;
    }
  gr[graphsize-1].reality = graphsize-2;
  gr[graphsize-1].label = graphsize-1;


  gr[0].desire = nodeLoc[1];                      //The first.
  gr[graphsize-1].desire = nodeLoc[graphsize-2];  //The last.
  for(unsigned i=1; i <= p.size(); i++)
    {
    gr[nodeLoc[(i*2)-1]].desire = nodeLoc[(i*2)-2];
    gr[nodeLoc[i*2]].desire = nodeLoc[(i*2)+1];
    }

  vector<bool> orientations = getRealityEdgeOrientations(p);
  for(unsigned i=0; i < 2*p.size()+2; i+=2)
    if(orientations[i/2])
      {
      gr[i].desire = i+1;
      gr[i+1].desire = i;
      }

  int n=gr[0].desire;  //The current node we're looking at.
  int label=1;         //The current label we going to use.
  while(n != 2*(int)p.size()+1)
    {
    p[n/2] = label++;
    n = gr[n+1].desire;
    }
  }





/*Get a component that would be broken by the given reversal.
  ** assume the only components are the unoriented ones.
*/
ipair Permutation::getBrokenComp(ipair r)
  {
  vector<ipair> ucomps = getUnorientedComps(getSeqVec());
  /*
  for(vector<ipair>::iterator c=ucomps.begin(); c!=ucomps.end(); c++)
    {                        //Fix those indices that run off the end.
    if(c->first == 0)
      c->first = 1;
    if(c->second == length()+1)
      c->second = length()+1;
    }
  */

  for(vector<ipair>::iterator c=ucomps.begin(); c!=ucomps.end(); c++)
    if(intersects(*c, r))
      return *c;

  vector<ipair> nested;      //Those components that r is nested in.
  for(vector<ipair>::iterator c=ucomps.begin(); c!=ucomps.end(); c++)
    if(isnested(r, *c) && r != *c)
      nested.push_back(*c);

  int maxleft = nested[0].first;
  vector<ipair>::iterator innermost=nested.begin();
  vector<ipair>::iterator c=nested.begin();
  while(c!=nested.end())
    {
    if(maxleft < c->first)
      {
      maxleft = c->first;
      innermost = c;
      }
    c++;
    }

  return *innermost;
  }


/* Return true if q is nested in p.
*/
inline bool Permutation::isnested(ipair q, ipair p)
  { return (q.first >= p.first && q.second <= p.second); }

inline bool Permutation::intersects(ipair p, ipair q)
  { return !(q.second < p.first || q.first > p.second ||
             isnested(q,p) || isnested(p,q)); }


/* Back out from a sequence of inversions until the most recent unsafe
inversion is met.
Return the adjacencies that would be undone from the identity for the set of
reversals that we've visited.
 ** The front-most element returned is the unsafe reversal that was done rather
    than an adjacency!
*/
#define DEF_T -1
list<ipair> Permutation::getUnstuck(Permutation &op, list<ipair> &revs)
  {
  list<ipair> undone;            //The reversals we will undo.
  vector<ipair> ucomps;
  if((ucomps = getUnorientedComps(getSeqVec())).size())
    {                                          //If we have done unsafe revs:
    Permutation redges(length()+1);            //keep the reality edges,

         //and set the component numbers for the reality edges:
    vector<BinaryNode*> nodes;                 //Get a vector of pointers to
    nodes = getPermPointers();                 //this in permutation order.
    vector<BinaryNode*> res;                   //Get a vector of pointers
    res = redges.getPermPointers();            //in permutation order.

    for(unsigned i=0; i<res.size(); i++)       //Innitialize the values:
      res[i]->clab = 0;
    
    vector<int> epnum(res.size());             //Endpoint of bad comp marker.
    for(unsigned h=0; h < ucomps.size(); h++)  //Mark the epnum endpoints:
      epnum[ucomps[h].first] = epnum[ucomps[h].second-1] = h+1;

                                               //Now fill in non-endpoints:
    stack<int> epnums;       //Stack to hold the bad comp endpoint numbers.
    stack<int> blabs;        //Stack to hold the nested bad comp labels.
    vector<int> tlabs;       //Stack for the trivial edge group labels.
                             //(trivial edges under a comp must have different
                             // labels if seperated by non-bps in that comp)
    vector<bool> truns;      //True if last thing seen was a trivial edge.
    int newlabel = 0;        //The label to mark the next component with.

    epnums.push(0);          //Push the non-unoriented comp value on the stack.
    blabs.push(-1);          //It will never be used so add garbage.
    tlabs.push_back(DEF_T);
    truns.push_back(false);
    unsigned n = 1;
    res[0]->clab = newlabel++;     //Assign the label 0 to the first edge.
    if(epnum[0] > 0)               //If this node is a left bad comp endpoint
      {
      res[0]->bp = true;           //mark that it's a breakpoint,
      epnums.push(epnum[0]);       //push the epnum,
      blabs.push(0);               //push the bad component label,
      tlabs.push_back(DEF_T);      //push the null trivial edge label,
      truns.push_back(false);      //and mark that this edge is not trivial.
      }
    else                           //Otherwise, 
      {
      res[0]->bp = false;          //mark that it's not a breakpoint,
      tlabs.back() = 0;            //set the trivial edge label as using 0
      truns.back() = true;         //and mark the run of trivial edges.
      }
    while(n < res.size()-1)  //For each reality edge:
      {
      if(epnum[n] == 0)            //If this node is not a bad comp endpoint:
        {
        if(abs(nodes[n-1]->element)+1 == abs(nodes[n]->element))
          {                        //and it's trivial (non-breakpoint),
                                   //mark clab with the current bad comp num.
          if(!truns.back())        //if we're not on a run of trivial edges
            {                      //then mark the new label
            tlabs.back() = newlabel++;
            truns.back() = true;   //and mark that we are on a run.
            }                      //Label the edge with current trivial label.
          res[n]->clab = tlabs.back();
          res[n]->bp = false;
          }
        else                       //Otherwise we are at a breakpoint inside
          {                        //a bad component
                                   //so mark clab with the current badcomp label
          res[n]->clab = blabs.top();
          truns.back() = false;    //and record that ther's no trivial edge run.
          res[n]->bp = true;
          }
        }
      else if(epnum[n] == epnums.top())
        {                       //It's bad comp and the endpoint is the current:
                                // (i.e. we're at the right endpoint)
        res[n]->clab = blabs.top();
        epnums.pop();
        blabs.pop();
        tlabs.pop_back();
        truns.pop_back();
        res[n]->bp = true;
        }
      else                      //We have a new left endpoint
        {                       //so mark with the new label:
        res[n]->clab = newlabel++;
        epnums.push(epnum[n]);
        blabs.push(newlabel-1);
        tlabs.push_back(DEF_T);
        truns.push_back(false);
        res[n]->bp = true;
        }
      n++;
      }
    if(abs(nodes.back()->element) == (int)nodes.size())
      {                                  //If the last edge is a non-breakpoint
      res.back()->bp = false;
      if(truns.back())                   //and we're on a run
        res.back()->clab = tlabs.back(); //otherwise give the old label.
      else
        res.back()->clab = newlabel++;   //then give a new label,
      }
    else
      {
      res.back()->clab = blabs.top();    //Label the breakpoint with old label.
      res.back()->bp = true;
      }


    while(redges.isSafe(revs.back()))
      {
      ipair r = revs.back();
      rev(r);                                //Undo the rev on the permutation.
      op.rev(r);
      redges.rev(r.first, r.second+1);       //Do the rev on the reality edges.

      undone.push_front(op.getAdjCreated(r));//Record the adjacency created.
      revs.pop_back();
      }

    op.rev(revs.back());
    undone.push_front(revs.back());          //Save the unsafe reversal.
    revs.pop_back();
    }

  return undone;
  }


/* Return the elements of the adjacency created for -r- on this permutation.
   Say the reversal looks like    oi ri...rj oj =>  oi -rj...-ri oj.
*/
ipair Permutation::getAdjCreated(ipair r)
  {
  int oi,oj;                   //Get the elements being acted upon
  if(r.first == 1)
    oi = 0;
  else
    {
    splay(r.first-2, root);
    oi = root->element;
    }
  splay(r.first-1, root);
  int ri = -root->element;     //but negate the reversed elements immediately.
  splay(r.second-1, root);
  int rj = -root->element;
  if(r.second == length())
    oj = length()+1;
  else
    {
    splay(r.second, root);
    oj = root->element;
    }


  if(oi-rj == -1)              //If we made an adjacency on the left
    {
    if(oi >= 0)                //and the elements are positive
      return ipair(oi,rj);     //then return them in order.
    else
      return ipair(-rj,-oi);   //Otherwise, return the negatives.
    }
  if(ri-oj == -1)              //If we made an adjacency on the right
    {
    if(oj > 0)                 //and the elements are positive
      return ipair(ri,oj);     //then return them in order.
    else
      return ipair(-oj,-ri);   //Otherwise, return the negatives.
    }

  cerr << "Permutation::getAdjCreated: SHOULD NEVER GET HERE!" << endl;
  exit(0);
  }


/* Return the corresponding reversal on the inverse permutation.
*/
ipair Permutation::findInvRev(ipair r)
  {
  int lendpoint;
  int rendpoint;

  splay(r.first-1, root);
  BinaryNode* left = root;
  splay(r.second-1, root);
  BinaryNode* right = root;

  updateNodeInfo(left);                //Make sure the signs on the
  updateNodeInfo(right);               //endpoints are correct.
  if(abs(left->element) < abs(right->element))
    {
    if(left->element > 0)
      lendpoint = abs(left->element);
    else
      lendpoint = abs(left->element)+1;
    if(right->element > 0)
      rendpoint = abs(right->element);
    else
      rendpoint = abs(right->element)-1;
    }
  else
    {
    if(left->element > 0)
      rendpoint = abs(left->element)-1;
    else
      rendpoint = abs(left->element);
    if(right->element > 0)
      lendpoint = abs(right->element)+1;
    else
      lendpoint = abs(right->element);
    }

  return ipair(lendpoint,rendpoint);
  }




bool Permutation::hasHurdle() const
  { return getUnorientedComps(getSeqVec()).size(); }

/*
Input: Permutation of n-1 elements framed by 0 and n
Output: Indices of endpoints of each hurdle in the permutation or empty vector if there is no hurdle
typedef enum {PLUS, MINUS} sign;
vector<ipair> Permutation::getUnorientedComps() const
{
  vector<int> p = getFramedSeqVec();

  stack<int> s1;
  stack<int> s2;
  stack<int> m1;
  stack<int> m2;

  vector<ipair> hurdleList;

  int n = p.size();
  int *max = new int[n];
  int *min = new int[n];
  int *M = new int[n];
  int *m = new int[n];
  bool *mark1 = new bool[n];
  bool *mark2 = new bool[n];
  int *pi = new int[n];
  sign *si = new sign[n];

  for(int i=0; i < n; i++)
  {
    if(p.at(i) > 0)
    {
      pi[i] = p.at(i);
      si[i] = PLUS;
    }
    else
    {
      pi[i] = 0 - p.at(i);
      si[i] = MINUS;
    }
  }

  M[0] = n-1;
  m[0] = 0;
  max[0] = 0;
  min[0] = n-1;
  mark1[0] = false;
  mark2[0] = false;
  s1.push(0);
  s2.push(0);
  m1.push(n);
  m2.push(0);

  int t;

  for(int i = 1; i < n; i++)
  {

    //compute M[i]
    if (pi[i-1] > pi[i])
    {
      M[i] = pi[i-1];
      m1.push(pi[i-1]);
    }
    else
    {
      while(m1.top() < pi[i])
        m1.pop();
      M[i] = m1.top();
    } 
    //find connected components
    while(pi[i] < pi[(t = s1.top())] || pi[i] > M[t])
    {
      s1.pop();
      max[s1.top()] = (max[s1.top()] > max[t])? max[s1.top()] : max[t];
      mark1[s1.top()] |= mark1[t];
    }
    if((si[i] == PLUS) && (pi[i] == max[t = s1.top()]+1) &&
       (i-t == pi[i] - pi[t]) && (i-t > 1))
    {
      if(mark1[t] == false)
      {
        ipair hurdle;
        hurdle.first = t;
        hurdle.second = i;
        assert(t < i);
        hurdleList.push_back(hurdle);
      }
      mark1[s1.top()] = false;
    }

    //compute m[i]
    if (pi[i-1] < pi[i])
    {
      m[i] = pi[i-1];
      m2.push(pi[i-1]);
    }
    else
    {
      while(m2.top() > pi[i])
        m2.pop();
      m[i] = m2.top();
    } 
    //find connected components
    while((pi[i] > pi[(t = s2.top())] || pi[i] < m[t]) && t>0)
    {
      s2.pop();
      min[s2.top()] = (min[s2.top()] < min[t])? min[s2.top()] : min[t];
      mark2[s2.top()] |= mark2[t];
    }
    if((si[i] == MINUS) && (pi[i] == min[t = s2.top()]-1) &&
       (i-t == pi[t] - pi[i]) && (i-t > 1))
    {
      if(mark2[t] == false)
      {
        ipair hurdle;
        hurdle.first = t;
        hurdle.second = i;
        assert(t < i);
        hurdleList.push_back(hurdle);
      }
      mark2[s2.top()] = false;
    }

    if (si[i] == PLUS)
      s1.push(i);
    else
      s2.push(i);
    max[i] = pi[i];
    min[i] = pi[i];
    min[s2.top()] = (min[s2.top()] < pi[i]) ? min[s2.top()] : pi[i];
    max[s1.top()] = (max[s1.top()] > pi[i]) ? max[s1.top()] : pi[i];
    mark1[s1.top()] = si[i]==MINUS;
    mark2[s2.top()] = si[i]==PLUS;
  }

  delete []max;
  delete []min;
  delete []M;
  delete []m;
  delete []mark1;
  delete []mark2;
  delete []pi;
  delete []si;

  return hurdleList;
}
*/

// Apply the given reversals to the permutation.
void Permutation::applyRevsLastTOFirst(list<ipair> revs)
  {
  for(list<ipair>::reverse_iterator r=revs.rbegin(); r!=revs.rend(); r++)
    { rev(*r); }
  }

/* Return true if the reversal given would result in the same number of
unoriented components.
    ** this Permutation must represent the reality edges and have clab
       and bp labelled properly!

If either of the given endpoints are a breakpoint then we know it's not safe.
If they both are not breakpoints but span only part of a component then it's
not safe.
Otherwise, it is safe.
*/
bool Permutation::isSafe(const ipair i1)
  {
  splay(i1.first-1, root);
  BinaryNode *i = root;
  splay(i1.second, root);
  BinaryNode *j = root;

  return !(i->bp || j->bp || (i->clab != j->clab));
  }


/* Update the given enterval indices if it is a subinterval of the -rev-.
*/
void Permutation::updateInterval(const ipair rev, ipair &interval) const
  {
  if(interval.first > rev.first && interval.second < rev.second)
    {
    int change = rev.second - interval.second;
    interval.first = interval.second-interval.first + change;
    interval.second = interval.first + interval.second-interval.first;
    }
  }



/* Do the next sorting reversal (return false otherwise).
*/
bool Permutation::nextRev(int flag)
  {
  ipair p;
  return nextRev(p, flag);
  }
bool Permutation::nextRev(ipair &retval, int flag)
  {
  retval.first = retval.second = 0;
  updateFields(root);
  
  if(flag == RANDOM)            //assign flag randomly
    flag = random()%2;

  if(isAllPositive())
    return false;

  retval = doOrientedRev(flag);

#ifdef TRY_OTHER_ON_FAILURE
  if(isAllPositive())           //check if minNeg == INT_MAX
    {
    rev(retval);                       //Undo the last one
    retval = doOrientedRev(!flag);     //and redo the other.
    }
#endif

  return true;
  }






/* Do the oriented reversal by the method in -flag-.  Return the reversal.
*/
ipair Permutation::doOrientedRev(int flag)
  {
  ipair retval;

    //Find the first endpoint:
  BinaryNode* i = root;
  for(;;)
    {
    updateFields(i);
    if(!stepDown(i, flag)) //Move towards the min negative or max positive and
      break;               //break when we reach that node.
    }
  retval = getOtherEndpoint(i, flag);
  rev(retval);
  return retval;
  }



/*Return the endpoints given by the method in -flag-.
  If flag is USE_MIN then the positive element less than i->element is found
  as the other endpoint. Otherwise, the one greater is used.
*/
ipair Permutation::getOtherEndpoint(BinaryNode* i, int flag)
  {
  ipair r = NULL_IPAIR;

  splay(i->index,root);                       //Not necessary?!!!!!!!!!!

    //Get the second endpoint:
  BinaryNode* j;
  ipair adj;
  if(isEndElement(i, r, flag))
    {
/*
    if(abs(i->element) == 1)
      adj = ipair(0,1);
    else
      adj = ipair(length(),length()+1);
*/
    }
  else
    {
    j = getOtherEnd((i->element), flag);
    updateNodeInfo(j);

    BinaryNode* tmp;
    if(i->index > j->index)
      {
      tmp = i;
      i = j;
      j = tmp;
      }
    fixIndices(r, i->index, j->index, flag);

/*
    if(abs(i->element) < abs(j->element))     //Get the adjacency created:
      adj = ipair(abs(i->element),abs(j->element));
    else
      adj = ipair(abs(j->element),abs(i->element));
*/
    }


  //return pair<ipair,ipair>(r,adj);
  return r;
  }


/* Assure that the given node has its index and sign updated.
*/
void Permutation::updateNodeInfo(BinaryNode* j)
  {
    //Head to the root or the other one and then fix the reversal flags:
  list<BinaryNode *> path;
  BinaryNode* tmp = j;
  while(tmp->parent != nullNode && tmp->parent != tmp)
    {
    path.push_front(tmp);
    tmp = tmp->parent;
    }
  path.push_front(tmp);
  
  for(list<BinaryNode*>::iterator k=path.begin(); k!=path.end(); k++)
    {
    pushRevDown(*k);
    pushIndDown(*k);
    }
  }


/* Return the length of the permutation.
*/
inline int Permutation::length()
  { return root->size; }


inline bool Permutation::isAllPositive()
{
  if(root->maxNegative == 0)   //If there is no negative value then
    return true;               //return.
  else
    return false;
}

//go down the tree in the direction of minNeg or maxNeg, modifies i
//break if we reach the node with minNeg/maxNeg value  
inline bool Permutation::stepDown(BinaryNode* &i, int flag)
{
  if(flag == USE_MIN) //minNegative
  {
    int best = i->minNegative;
    if(i->left->minNegative == best)
      i = i->left;
    else if(i->right->minNegative == best)
      i = i->right;
    else
      return false;
  }
  else if (flag == USE_MAX)
  {
    int best = i->maxNegative;
    if(i->left->maxNegative == best)
      i = i->left;
    else if(i->right->maxNegative == best)
      i = i->right;
    else
      return false;
  }
  return true;
}
 
inline bool Permutation::isEndElement(BinaryNode* i, ipair &retval, int flag)
{
  if (flag == USE_MIN) //minNegative
  {  
    if(abs(i->element) == 1)
    {
    retval.first = 1;
    retval.second = i->index+1;
    }
    else
      return false;
  }
  else if (flag == USE_MAX) //maxNegative
  {
    if(abs(i->element) == root->size)
    {
    retval.first = i->index+1;
    retval.second = root->size;
    }
    else
      return false;
  }
  return true;
}

inline Permutation::BinaryNode* Permutation::getOtherEnd(int elem, int flag)
{
  if (flag == USE_MIN) //minNegative
    return lookup[abs(elem)-1];

  return lookup[abs(elem)+1];
}

inline void Permutation::fixIndices(ipair &retval, int i, int j, int flag)
{
  if (flag == USE_MIN) //minNegative
  {
    retval.first = i+2;
    retval.second =  j+1;
  }
  else if (flag == USE_MAX) //maxNegative
  {
    retval.first = i+1;
    retval.second =  j;
  }
}

/* Reverse the permutation over the given indices.
   p1...p(i-1) p(i)... p(j)p(j+1)...pn becomes
   p1...p(i-1)-p(j)...-p(i)p(j+1)...pn
*/
//assume: i, j \in [1,n], j >= i, n = root->size
void Permutation::rev(ipair r)
  { rev(r.first, r.second); }
void Permutation::rev(int i, int j)
{
    //indices in tree: 0...(n-1)
 
    if(i <= 1) { //need only 1 split
	BinaryNode *rhs = split(j-1, root);
	root->reversed = true;
	root->stale = true;
	join(root, rhs);
    }
    else {
        BinaryNode *mid = split(i-2, root);
        BinaryNode *rhs = split(j-1, mid);
	mid->reversed = true;
	mid->stale = true;
        join(root,mid);         //Be carefull to join them in this order so
        join(root,rhs);         //that the indices can be maintained.
    }
    //updateFields(root);
}






/* Get a reversal that we know is safe, given the unsafe reversal -r-.
   assume: the hurdle created by -r- contains only positive elements
*/
ipair Permutation::getSafeRev(ipair reversal, ipair interval) const
{
  vector<int> p = getFramedSeqVec();
  ipair safeRev;

  //convert to zero-start-index
  int ri = reversal.first, rj = reversal.second;
  int hi = interval.first, hj = interval.second;

  assert((ri != hi) || (rj != hj));

  int num = getUnorientedComps(getSeqVec()).size();
  vector<int> index (p.size());

  for(int i = 0; i < (int)p.size(); i++)
    index[abs(p[i])] = i;
  
  int i,less, more;
  //reversal is contained in the interval
  if((ri > hi) && (rj < hj))
    {
      for(i = ri; i <= rj; i++)
      {
         less = index[abs(p[i])-1];
         more = index[abs(p[i])+1];
         if(p[i] < 0)
         {
         if(less < ri && less >= hi && -p[i]-p[less]==1)
         {
           if(less+1 != ri || i != rj)
           {
             safeRev.first = less+1;
             safeRev.second = i;
             break;
           }
         }
         if(more < ri && more >= hi && p[i]+p[more] == 1)
         {
           if(more != ri || i-1 != rj)
           {
             safeRev.first = more;
             safeRev.second = i-1;
             break;
           }
         }
         if(less > rj && less <= hj && -p[less]-p[i] == 1)
         {
           if(i+1 != ri || less != rj)
           {
             safeRev.first = i+1;
             safeRev.second = less;
             break;
           }
         }
         if(more > rj && more <= hj && p[more]+p[i] == 1)
         {
           if(i != ri || more-1 != rj)
           {
             safeRev.first = i;
             safeRev.second = more-1;
             break;
           }
         }
         }
         else
         {
         if(less < ri && less >= hi && p[i]+p[less]==1)
         {
           if(less+1 != ri || i != rj)
           {
             safeRev.first = less;
             safeRev.second = i-1;
             break;
           }
         }
         if(more < ri && more >= hi && -p[i]-p[more] == 1)
         {
           if(more != ri || i-1 != rj)
           {
             safeRev.first = more+1;
             safeRev.second = i;
             break;
           }
         }
         if(less > rj && less <= hj && p[less]+p[i] == 1)
         {
           if(i+1 != ri || less != rj)
           {
             safeRev.first = i;
             safeRev.second = less-1;
             break;
           }
         }
         if(more > rj && more <= hj && -p[more]-p[i] == 1)
         {
           if(i != ri || more-1 != rj)
           {
             safeRev.first = i+1;
             safeRev.second = more;
             break;
           }
         }
         }
      }
      Permutation q(getSeqVec());
      q.rev(safeRev.first, safeRev.second);
      if(num < (int)getUnorientedComps(q.getSeqVec()).size())
        safeRev = getSafeRev(safeRev, q.getBrokenComp(safeRev));     
    }
  //reversal intersects from the right
  else if((rj > hj) || ((ri > hi) && (rj == hj)))
    {
      //left = rj-hj+ri;
      //right = rj;
      for(i = rj-hj+ri; i <= rj; i++)
      {
         less = index[abs(p[i])-1];
         more = index[abs(p[i])+1];
         if(p[i] < 0)
         {
         if(less < ri && less >= hi && -p[i]-p[less] == 1)
         {
           if(less+1 != ri || i != rj)
           {
             safeRev.first = less+1;
             safeRev.second = i;
             break;
           }
         }
         if(more < ri && more >= hi && p[i] + p[more] == 1)
         {
           if(more != ri || i-1 != rj)
           {
             safeRev.first = more;
             safeRev.second = i-1;
             break;
           }
         }
         }
         else
         {
         if(less < ri && less >= hi && p[i]+p[less] == 1)
         {
           if(less+1 != ri || i != rj)
           {
             safeRev.first = less;
             safeRev.second = i-1;
             break;
           }
         }
         if(more < ri && more >= hi && -p[i] - p[more] == 1)
         {
           if(more != ri || i-1 != rj)
           {
             safeRev.first = more+1;
             safeRev.second = i;
             break;
           }
         }
         }
      }     
    }
  //reversal intersects from the left
  else if((ri < hi) || ((ri == hi) && (rj < hj)))
    {
      //left = ri;
      //right = rj-hi+ri;
      for(i = ri; i <= rj-hi+ri; i++)
      {
         less = index[abs(p[i])-1];
         more = index[abs(p[i])+1];
         if(p[i] < 0)
         {
         if(less > rj && less <= hj && -p[less]-p[i] == 1)
         {
           if(i+1 != ri || less != rj)
           {
             safeRev.first = i+1;
             safeRev.second = less;
             break;
           }
         }
         if(more > rj && more <= hj && p[more] + p[i] == 1)
         {
           if(i != ri || more-1 != rj)
           {
             safeRev.first = i;
             safeRev.second = more-1;
             break;
           }
         }
         }
         else{
         if(less > rj && less <= hj && p[less]+p[i] == 1)
         {
           if(i+1 != ri || less != rj)
           {
             safeRev.first = i;
             safeRev.second = less-1;
             break;
           }
         }
         if(more > rj && more <= hj && -p[more] - p[i] == 1)
         {
           if(i != ri || more-1 != rj)
           {
             safeRev.first = i+1;
             safeRev.second = more;
             break;
           }
         }
         }
      }     
    }
  else
    cerr << "whats this?!: r(" << ri << "," << rj << ") interval: (" << hi << "," << hj << ")" << endl;
  return safeRev;
}




/* Get a reversal that we know is safe, given the unsafe reversal -r- and
   the interval of the created hurdle -interval-.
   assume: the hurdle created by -r- contains only positive elements
*/
ipair Permutation::getSafeRevDumb(ipair r, ipair interval) const
  {
  vector<int> p = getFramedSeqVec();
  int j;
  for(int i = r.first; i <= r.second; i++)
    {
    if((j = getPositiveInd(p,(-p[i])-1,ipair(0,r.first-1))) >= 0 &&
       ipair(j+1,i) != r)                             //Less is to the left.
      return ipair(j+1,i);
    if((j = getPositiveInd(p,(-p[i])+1,ipair(0,r.first-1))) >= 0 &&
       ipair(j,i-1) != r)                             //More is to the left.
      return ipair(j,i-1);
    if((j = getPositiveInd(p,(-p[i])-1,ipair(r.second+1,p.size()-1))) >= 0 &&
       ipair(i+1,j) != r)                             //Less is to the right.
      return ipair(i+1,j);
    if((j = getPositiveInd(p,(-p[i])+1,ipair(r.second+1,p.size()-1))) >= 0 &&
       ipair(i,j-1) != r)                             //More is to the right.
      return ipair(i,j-1);
    }
  cerr << "Permutation::getSafeRevDumb: SHOULD NEVER GET HERE!" << endl;
  exit(0);
  }
/*
pair<ipair,ipair> Permutation::getSafeRevDumb(ipair r, ipair interval) const
  {
  vector<int> p = getFramedSeqVec();
  int j;
  for(int i = r.first; i <= r.second; i++)
    {
    if((j = getPositiveInd(p,(-p[i])-1,ipair(1,r.first-1))) &&
       ipair(j+1,i) != r)                             //Less is to the left.
      return pair<ipair,ipair>(ipair(j+1,i),ipair(-p[i]-1,-p[i]));
    if((j = getPositiveInd(p,(-p[i])+1,ipair(1,r.first-1))) &&
       ipair(j,i-1) != r)                             //More is to the left.
      return pair<ipair,ipair>(ipair(j,i-1),ipair(-p[i],-p[i]+1));
    if((j = getPositiveInd(p,(-p[i])-1,ipair(r.second+1,p.size()-1))) &&
       ipair(i+1,j) != r)                             //Less is to the right.
      return pair<ipair,ipair>(ipair(i+1,j),ipair(-p[i]-1,-p[i]));
    if((j = getPositiveInd(p,(-p[i])+1,ipair(r.second+1,p.size()-1))) &&
       ipair(i,j-1) != r)                             //More is to the right.
      return pair<ipair,ipair>(ipair(i,j-1),ipair(-p[i],-p[i]+1));
    }
  cerr << "Permutation::getSafeRevDumb: SHOULD NEVER GET HERE!" << endl;
  exit(0);
  }
*/

/* Return the index of the element in the given permutation if it is in the
given interval.  Otherwise, return -1.
*/
int Permutation::getPositiveInd(vector<int> p, int elem, ipair interval) const
  {
  for(int i=interval.first; i <= interval.second; i++)
    if(p[i] == elem)
      return i;
  return -1;
  }






/////////////////////////////////////////////////////////////
/**
 * Find the smallest item in the tree.
 * Not the most efficient implementation (uses two passes), but has correct
 *     amortized behavior.
 * A good alternative is to first call find with parameter
 *     smaller than any item in the tree, then call findMin.
 * Return the smallest item or throw UnderflowException if empty.
 */

Permutation::BinaryNode* Permutation::findMinP( )
{
    if( isEmpty( ) )
        throw UnderflowException( );

    splay( 0, root );
    return root;
}

int Permutation::findMin( )
  { return (*findMinP()).element; }

/**
 * Find the largest item in the tree.
 * Not the most efficient implementation (uses two passes), but has correct
 *     amortized behavior.
 * A good alternative is to first call find with parameter
 *     larger than any item in the tree, then call findMax.
 * Return the largest item or throw UnderflowException if empty.
 */

Permutation::BinaryNode* Permutation::findMaxP( )
{
    if( isEmpty( ) )
        throw UnderflowException( );

    splay( root->size-1, root );
    return root;
}

int Permutation::findMax( )
  { return (*findMaxP()).element; }

////////////////////////////////////////////////////
   
bool Permutation::contains( int index )
{
    if( isEmpty( ) )
        return false;
    splay( index, root );
    return root->index == index;
}



bool Permutation::isEmpty( ) const
{
    return root == nullNode;
}


vector<int> Permutation::getSeqVec( ) const
  { return getSeqVec(root); }


/* Get the sequence vector with frame elements.
*/
vector<int> Permutation::getFramedSeqVec() const
  {
  vector<int> p;
  p.push_back(0);
  vector<int> sv = getSeqVec(root);
  std::copy(sv.begin(), sv.end(), back_inserter(p));
  p.push_back(p.size());
  return p;
  }



void Permutation::printPerm( ) const
{
    if( isEmpty( ) )
        cerr << "Empty tree";
    else
        cerr << getSeqVec( );
    cerr << endl;
}

void Permutation::printNH() const
  {
  if(!isEmpty())
    printNH(root);
  cerr << ";" << endl;
  }



void Permutation::printCompInfo( ) const
{
    if( isEmpty( ) )
        cerr << "Empty tree";
    else
        printCompInfo( root );
    cerr << endl;
}

void Permutation::printFlags( ) const
{
    if( isEmpty( ) )
        cerr << "Empty tree";
    else
        printFlags( root );
    cerr << endl;
}


/* Return true if the permutation is the identity.
*/
bool Permutation::isSorted()
  {
  vector<int> perm = getSeqVec();
  return perm == makeIDv(length());
  }


//////////////////////////////////////////////////

/*
void Permutation::makeEmptySlow( )
{
// *****************************
// * Comment this out, because it is prone to excessive
// * recursion on degenerate trees. Use alternate algorithm.
    
//    reclaimMemory( root );
//    root = nullNode;
// *******************************
    while( !isEmpty( ) )
    {
        findMax( );        // Splay max item to root
        remove( root->index );
    }
    if( nullNode != NULL )
        delete nullNode;
}
*/


/* Remove all the nodes from the tree by using the -lookup- array.
*/
void Permutation::makeEmpty( )
  {
  delete [] mem;

  if(nullNode != NULL)
    delete nullNode;
  }
/*
void Permutation::makeEmpty( )
  {
  for(vector<BinaryNode*>::iterator i=lookup.begin(); i!=lookup.end(); i++)
    delete *i;

  if(nullNode != NULL)
    delete nullNode;
  }
*/

/////////////////////////////////////////////////////


/* Insert a node with given index using the given memory from -mem-.
*/
void Permutation::insert( int index, int x, int &j)
{
    static BinaryNode *newNode = NULL;

    if( newNode == NULL ){
        //newNode = new BinaryNode();
        newNode = &(mem[j]);
        j++;
    }
    newNode->element = x;
    newNode->index = index;

    if( root == nullNode )
    {
        newNode->left = newNode->right = newNode->parent = nullNode;
        root = newNode;
        root->size = 1;
        if (x < 0){
            root->minNegative = root->maxNegative = -x;
            root->maxPositive = root->minPositive = PHI;
        }
        else{
            root->minNegative = root->maxNegative = PHI;
            root->maxPositive = root->minPositive = x;
        }        
    }
    else
    {
        splay( index, root );
        if( index < root->index )
        {
            newNode->left = root->left;
            newNode->left->parent = newNode;

            newNode->right = root;
            root->parent = newNode;
            root->left = nullNode;

            updateFields(root);
            root = newNode;
            updateFields(root);
        }
        else
        if( root->index < index )
        {
            newNode->right = root->right;
            newNode->right->parent = newNode;

            newNode->parent = root->parent;
            newNode->left = root;
            root->parent = newNode;
            root->right = nullNode;

            updateFields(root);
            root = newNode;
            updateFields(root);
    
        }
        else
            return;
    }
    lookup[abs(x)] = newNode;
    newNode = NULL;   // So next insert will call new
}



/*
void Permutation::insert( int index, int x)
{
    static BinaryNode *newNode = NULL;

    if( newNode == NULL )
        newNode = new BinaryNode();
    newNode->element = x;
    newNode->index = index;

    if( root == nullNode )
    {
        newNode->left = newNode->right = newNode->parent = nullNode;
        root = newNode;
        root->size = 1;
        if (x < 0){
            root->minNegative = root->maxNegative = -x;
            root->maxPositive = root->minPositive = PHI;
        }
        else{
            root->minNegative = root->maxNegative = PHI;
            root->maxPositive = root->minPositive = x;
        }        
    }
    else
    {
        splay( index, root );
        if( index < root->index )
        {
            newNode->left = root->left;
            newNode->left->parent = newNode;

            newNode->right = root;
            root->parent = newNode;
            root->left = nullNode;

            updateFields(root);
            root = newNode;
            updateFields(root);
        }
        else
        if( root->index < index )
        {
            newNode->right = root->right;
            newNode->right->parent = newNode;

            newNode->parent = root->parent;
            newNode->left = root;
            root->parent = newNode;
            root->right = nullNode;

            updateFields(root);
            root = newNode;
            updateFields(root);
    
        }
        else
            return;
    }
    lookup[abs(x)] = newNode;
    newNode = NULL;   // So next insert will call new
}
*/

void Permutation::remove( int index )
{
    BinaryNode *newTree;

        // If x is found, it will be at the root
    splay( index, root );
    if( root->index != index )
        return;   // Item not found; do nothing

    if( root->left == nullNode ){
        newTree = root->right;
        newTree->parent = nullNode;
        newTree->stale = true;
    }        
    else
    {
        // Find the maximum in the left subtree
        // Splay it to the root; and then attach right child
        newTree = root->left;
        splay( index, newTree );
        newTree->right = root->right;
        newTree->right->stale = true;
    }
    delete root;
    root = newTree;
    updateFields(root);
}

/////////////////////////////////////////////////////
/*
    const Permutation & operator=( const Permutation & rhs )
    {
        if( this != &rhs )
        {
            makeEmpty( );
            root = clone( rhs.root );
        }

        return *this;
    }

*/
/////////////////////////////////////////////////////



/**
 * Internal method to reclaim internal nodes in subtree t.
 * WARNING: This is prone to running out of stack space.
 */
void Permutation::reclaimMemory( BinaryNode *t )
{
    if( t != t->left )
    {
        reclaimMemory( t->left );
        reclaimMemory( t->right );
        delete t;
    }
}
    
/**
* Internal method to print a subtree t in new hampshire format.
* WARNING: This is prone to running out of stack space.
*/
void Permutation::printNH(BinaryNode *t, bool rev) const
  {
  bool hasleft = t->left != t->left->left;
  bool hasright = t->right != t->right->left;
  
  rev = (rev != t->reversed);
  if(!hasleft && !hasright)          //If we're at a leaf
    {
    if(rev)                          //and we have met an odd number of revs
      cerr << -t->element;           //then print the label reversed.
    else                             //Otherwise
      cerr << t->element;            //just print the label.
    }
  else
    {
    cerr << "(";
  
    if(hasleft)                      //If our left child is not the nullNode
      printNH(t->left, rev);         //then explore it.
    if(hasleft && hasright)
      cerr << ",";
    if(hasright)                     //If our right child is not the nullNode
      printNH(t->right, rev);        //then explore it.
  
    cerr << ")" << t->element;
    }
  }

/**
 * Internal method to print a subtree t in sorted order.
 * WARNING: This is prone to running out of stack space.
 */
void Permutation::printFlags( BinaryNode *t ) const
{
    if( t != t->left )
    {
        printFlags( t->left );
        if( t->reversed )
          cerr << "r";
        if( t->stale )
          cerr << "s";
        cerr << t->element << "[" << t->index << "]" << t->size << ":"
             << t->minNegative << "<->" << t->minPositive << "  "
             << t->maxNegative << "<->" << t->maxPositive << endl;
        printFlags( t->right );
    }
}


void Permutation::printCompInfo( BinaryNode *t, bool rev) const
  {
  if( t != t->left )
    {
    rev = (rev != t->reversed);
    if(rev)
      {
      printCompInfo( t->right );
      cerr << abs(t->element) << "(";
      cerr << t->clab << "," << t->bp << ")";
      printCompInfo( t->left );
      }
    else
      {
      printCompInfo( t->left );
      cerr << abs(t->element) << "(";
      cerr << t->clab << "," << t->bp << ")";
      printCompInfo( t->right );
      }
    }
  }



void Permutation::printPerm( BinaryNode *t ) const
  { cerr << getSeqVec(t) << endl; }

/*
vector<int> Permutation::getSeqVec( BinaryNode *t ) const
  {
  vector<int> ret;
  getSeqVecRec(t, ret);
  return ret;
  }
*/

/**
 * Internal method to print a subtree t in sorted order.
 * WARNING: This is prone to running out of stack space.
 */
void Permutation::getSeqVecRec( BinaryNode *t, vector<int> &ret, bool rev ) const
{
    if( t != nullNode )
    {
        rev = (rev != t->reversed);
        int e;
        if(rev)
          e = -t->element;
        else
          e = t->element;
        if( rev )
        {
            getSeqVecRec( t->right, ret, rev );
            ret.push_back(e);
            getSeqVecRec( t->left, ret, rev );
        }
        else
        {
            getSeqVecRec( t->left, ret, rev );
            ret.push_back(e);
            getSeqVecRec( t->right, ret, rev );
        }
    }
}


/*Get a vector of the permutation through an iterative approach.
*/
vector<int> Permutation::getSeqVec(BinaryNode *t) const
  {
  vector<int> ret;
  bool rev = false;

  BinaryNode *endNode = t->parent; 
  BinaryNode *lastNode = t; 
  BinaryNode *thisNode = t; 
  while(thisNode != endNode)
    {
    if(lastNode == thisNode || lastNode == thisNode->parent)
      {                                 //If we're coming down from parent:
//cerr << "down:" << thisNode->element;
      rev = (rev != thisNode->reversed);       //flip the reverse flag
      BinaryNode *left, *right;
      if(rev)                           //correct for possible reversed order
        {
        left = thisNode->right;
        right = thisNode->left;
        }
      else
        {
        left = thisNode->left;
        right = thisNode->right;
        }

      if(left != nullNode)              //move to the left child if possible
        {
//cerr << " left" << endl;
        lastNode = thisNode;
        thisNode = left;
        }
      else if(right != nullNode)
        {                               //move to the right child if possible
//cerr << " right" << endl;
        if(rev)
          ret.push_back(-thisNode->element);
        else
          ret.push_back(thisNode->element);
        lastNode = thisNode;
        thisNode = right;
        }
      else                              //move back to the parent
        {
//cerr << " up" << endl;
        if(rev)
          ret.push_back(-thisNode->element);
        else
          ret.push_back(thisNode->element);
        rev = (rev != thisNode->reversed);     //undo the reversed state
        lastNode = thisNode;
        thisNode = thisNode->parent;
        }
      }
    else
      {
      BinaryNode *left, *right;
      if(rev)                           //correct for possible reversed order
        {
        left = thisNode->right;
        right = thisNode->left;
        }
      else
        {
        left = thisNode->left;
        right = thisNode->right;
        }

      if(lastNode == left)              //If we're coming from the left child:
        {
//cerr << "up left:" << thisNode->element << endl;
        if(rev)
          ret.push_back(-thisNode->element);
        else
          ret.push_back(thisNode->element);
        if(right != nullNode)
          {                             //move to the right child if possible
          lastNode = thisNode;
          thisNode = right;
          }
        else                            //move back to the parent
          {
          rev = (rev != thisNode->reversed); //undo the reversed state
          lastNode = thisNode;
          thisNode = thisNode->parent;
          }
        }
      else                                 //We're coming from the right child:
        {
//cerr << "up right:" << thisNode->element << endl;
        rev = (rev != thisNode->reversed); //undo the reversed state
        lastNode = thisNode;
        thisNode = thisNode->parent;
        }
      }
    }
/*
vector<int> recone;
getSeqVecRec(t, recone);
if(ret != recone)
  {
  printNH();
  printFlags();
  exit(1);
  }
*/

  return ret;
  }



/* Get a vector of pointers (to the nodes) in permutation order.
*/
vector<Permutation::BinaryNode*> Permutation::getPermPointers() const
  { return getPermPointers(root); }

/*
vector<Permutation::BinaryNode*> Permutation::getPermPointers() const
  {
  vector<BinaryNode*> ret;
  getPermPointers(root, ret);
  return ret;
  }
*/


vector<Permutation::BinaryNode*> Permutation::getPermPointers(BinaryNode *t) const
  {
  vector<BinaryNode*> ret;
  bool rev = false;

  BinaryNode *endNode = t->parent; 
  BinaryNode *lastNode = t; 
  BinaryNode *thisNode = t; 
  while(thisNode != endNode)
    {
    if(lastNode == thisNode || lastNode == thisNode->parent)
      {                                 //If we're coming down from parent:
      rev = (rev != thisNode->reversed);       //flip the reverse flag
      BinaryNode *left, *right;
      if(rev)                           //correct for possible reversed order
        {
        left = thisNode->right;
        right = thisNode->left;
        }
      else
        {
        left = thisNode->left;
        right = thisNode->right;
        }

      if(left != nullNode)              //move to the left child if possible
        {
        lastNode = thisNode;
        thisNode = left;
        }
      else if(right != nullNode)
        {                               //move to the right child if possible
        ret.push_back(thisNode);
        lastNode = thisNode;
        thisNode = right;
        }
      else                              //move back to the parent
        {
        ret.push_back(thisNode);
        rev = (rev != thisNode->reversed);     //undo the reversed state
        lastNode = thisNode;
        thisNode = thisNode->parent;
        }
      }
    else
      {
      BinaryNode *left, *right;
      if(rev)                           //correct for possible reversed order
        {
        left = thisNode->right;
        right = thisNode->left;
        }
      else
        {
        left = thisNode->left;
        right = thisNode->right;
        }

      if(lastNode == left)              //If we're coming from the left child:
        {
        ret.push_back(thisNode);
        if(right != nullNode)
          {                             //move to the right child if possible
          lastNode = thisNode;
          thisNode = right;
          }
        else                            //move back to the parent
          {
          rev = (rev != thisNode->reversed); //undo the reversed state
          lastNode = thisNode;
          thisNode = thisNode->parent;
          }
        }
      else                                 //We're coming from the right child:
        {
        rev = (rev != thisNode->reversed); //undo the reversed state
        lastNode = thisNode;
        thisNode = thisNode->parent;
        }
      }
    }

  return ret;
  }


void Permutation::getPermPointersRec(BinaryNode *t, vector<BinaryNode*> &ret,
                                     bool rev) const
{
    if( t != nullNode )
    {
        rev = (rev != t->reversed);
        if( rev )
        {
            getPermPointersRec( t->right, ret, rev );
            ret.push_back(t);
            getPermPointersRec( t->left, ret, rev );
        }
        else
        {
            getPermPointersRec( t->left, ret, rev );
            ret.push_back(t);
            getPermPointersRec( t->right, ret, rev );
        }
    }
}



/* Print the permutation ignoring the changes from the reversed flags.
*/
void Permutation::printWrongPerm( BinaryNode *t) const
{
    if( t != nullNode )
    {
        printWrongPerm( t->left );
        cerr << t->element << " ";
        printWrongPerm( t->right );
    }
}

/* Print the paths from every node to the root.
*/
void Permutation::printNodesToRoot() const
  {
  cerr << "Parent Paths:" << endl;
  vector<BinaryNode*>::const_iterator i=lookup.begin();
  for(i=++i; i != lookup.end(); i++)
    {
    BinaryNode* n = *i;
    while(n != nullNode)
      {
      cerr << n->element << " ";
      n = n->parent;
      }
    cerr << endl;
    }
  }



    /**
     * Internal method to clone subtree.
     * WARNING: This is prone to running out of stack space.
     */
/*
    BinaryNode * clone( BinaryNode *t ) const
    {
        if( t == t->left )  // Cannot test against nullNode!!!
            return nullNode;
        else
            return new BinaryNode( t->index, clone( t->left ), clone( t->right ), clone( t->parent ), t->element, t->reversed, t->stale );
    }
*/

    // Tree manipulations
void Permutation::rotateWithLeftChild( BinaryNode * & k2 )
{
//cerr << "rotateWithLeftChild()" << endl;
    BinaryNode *p = k2->parent;
    BinaryNode *k1 = k2->left;
    k2->left = k1->right;
    k2->left->parent = k2;

    k1->right = k2;
    k2->parent = k1;

    updateFields(k2);
    k2 = k1;
    k2->parent = p;
    updateFields(k2);
}

void Permutation::rotateWithRightChild( BinaryNode * & k1 )
{
//cerr << "rotateWithRightChild:" << k1->element << endl;
    BinaryNode *p = k1->parent;
    BinaryNode *k2 = k1->right;
    k1->right = k2->left;
    k1->right->parent = k1;

    k2->left = k1;
    k1->parent = k2;

    updateFields(k1);
    k1 = k2;
    k1->parent = p;
    updateFields(k1);
}

/**
 * Internal method to perform a top-down splay.
 * The last accessed node becomes the new root.
 * This method may be overridden to use a different
 * splaying algorithm, however, the splay tree code
 * depends on the accessed item going to the root.
 * x is the target index to splay around.
 * t is the root of the subtree to splay.
 */
void Permutation::splay( int index, BinaryNode * & t )
{
    BinaryNode *leftTreeMax, *rightTreeMin, *parent;
    static BinaryNode header;
    parent = t->parent;

    header.parent = nullNode;
    header.left = header.right = nullNode;
    leftTreeMax = rightTreeMin = &header;

    nullNode->index = index;   // Guarantee a match

//cerr << "splaying:" << index << " on:"; printNH(t); cerr << endl;
    for( ; ; )
    {                                 //If a rev has been done on the subtree,
        pushRevDown(t);               //then push it down to the children.
                                      //If the subtree's indices are stale, then
        pushIndDown(t);               //fix this index and push stale down.
        if( index < t->index )
        {                             //Push down the reversal flags for the
            pushRevDown(t->left);     //left child if necessary.
                                      //Push down the stale flags for the
            pushIndDown(t->left);     //left child if necessary.
            if( index < t->left->index )
            {
                rotateWithLeftChild( t );
                pushRevDown(t->left);
                pushIndDown(t->left);
            }
            if( t->left == nullNode )
                break;
            // Link Right
            rightTreeMin->left = t;
            t->parent = rightTreeMin;
            rightTreeMin = t;
            t = t->left;
            t->parent = nullNode;
        }
        else if( t->index < index )
        {                             //Push down the reversal flags for the
            pushRevDown(t->right);    //right child if necessary.
                                      //Push down the reversal flags for the
            pushIndDown(t->right);    //right child if necessary.
            if( t->right->index < index )
            {
                rotateWithRightChild( t );
                pushRevDown(t->right);
                pushIndDown(t->right);
            }
            if( t->right == nullNode )
                break;

            // Link Left
            leftTreeMax->right = t;
            t->parent = leftTreeMax;
            leftTreeMax = t;
            t = t->right;
            t->parent = nullNode;
        }
        else
            break;
    }

    bool updatelsubtree = false;
    if(header.right->size > 0)
        updatelsubtree = true;
    bool updatersubtree = false;
    if(header.left->size > 0)
        updatersubtree = true;


    leftTreeMax->right = t->left;
    if( updatelsubtree )
    {
        if(t->left != nullNode)
            t->left->parent = leftTreeMax;
        updateFieldsInPath(leftTreeMax);
    }

    rightTreeMin->left = t->right;
    if(updatersubtree)
    {
        if(t->right != nullNode)
            t->right->parent = rightTreeMin;
        updateFieldsInPath(rightTreeMin);
    }


    t->left = header.right;
    t->left->parent = t;
    t->right = header.left;
    t->right->parent = t;
    t->parent = parent;
    updateFields(t);
}

/* Fix the index of this node based on it's parent's index and the size of
the subtrees below it.
*/
inline void Permutation::pushIndDown(BinaryNode *t)
  {
  if(t->stale)                                       //If we're stale then
    {                                                //update the index:
    if(t == nullNode)     //If we're at nullNode then leave.
      return;
  
    if(t->parent == nullNode)                        //If we're at nullNode
      t->index = t->left->size;                      //then add the size of 
    else if(t->parent->right == t)                   //If we're the right child
      t->index = t->parent->index+1 + t->left->size; //then add the size of 
    else                                             //out left subtree.
      t->index = t->parent->index-1 - t->right->size;
    t->stale = false;
    t->left->stale = true;
    t->right->stale = true;
    }
  }
  

/* Swap the subtrees and flip their flags.
*/
inline void Permutation::pushRevDown(BinaryNode *t)
  {
  if(t->reversed)
    {
    BinaryNode* tmp;
    tmp = t->left;                            //Swap the subtrees,
    t->left = t->right;
    t->right = tmp;
  
    swapMM(t);                                //swap the negative and postive,
  
    t->left->reversed = !t->left->reversed;   //invert their flags,
    t->right->reversed = !t->right->reversed;
  
    t->element = 0 - t->element;              //switch the sign on the element,
    t->reversed = false;                      //and mark it as being done.
    }
  }

/* Swap the min and max holder for the node.
*/
void Permutation::swapMM(BinaryNode *t)
  {
  static int i;
  i = t->minNegative;
  t->minNegative = t->minPositive;
  t->minPositive = i;
  i = t->maxNegative;
  t->maxNegative = t->maxPositive;
  t->maxPositive = i;
  }




inline int Permutation::MAX_N (int root, int left, int right)
{
    int max = (left > right) ? left : right;
    if (root < 0){
        root = 0 - root;
        return ((root > max) ? root : max);
    }
    else 
        return max;
}
inline int Permutation::MAX_P (int root, int left, int right)
{
    int max = (left > right) ? left : right;
    if (root > 0)
        return ((root > max) ? root : max);
    else 
        return max;
}
inline int Permutation::MIN_N (int root, int left, int right)
{
    int min = (left < right) ? left : right;
    if (root < 0){
        root = 0 - root;
        return ((root < min) ? root : min);
        }
    else 
        return min;
}
inline int Permutation::MIN_P (int root, int left, int right)
{
    int min = (left < right) ? left : right;
    if (root > 0){
        return ((root < min) ? root : min);
        }
    else 
        return min;
}

inline void Permutation::updateFields(BinaryNode *t)
{
    t->size = t->left->size + t->right->size + 1;
    pushRevDown(t);
    pushIndDown(t);
    pushRevDown(t->left);
    pushIndDown(t->left);
    pushRevDown(t->right);
    pushIndDown(t->right);
    t->maxNegative = MAX_N(t->element, t->right->maxNegative, t->left->maxNegative);
    t->maxPositive = MAX_P(t->element, t->right->maxPositive, t->left->maxPositive);
    t->minPositive = MIN_P(t->element, t->right->minPositive, t->left->minPositive);
    t->minNegative = MIN_N(t->element, t->right->minNegative, t->left->minNegative);
}

void Permutation::updateFieldsInPath(BinaryNode *t)
{
    while (t != nullNode) {
        updateFields(t);
        t = t->parent;
        } 
}

/* Split the tree (rootel at -root-) at the given index.
Do this by splaying the index to the root and then chopping the right child of
the root.
*/
Permutation::BinaryNode * Permutation::split(int index, BinaryNode *&r)
{
    splay(index, r);        //Bring the element to the root so that we
                            //splay removed the reversed and stale flags from
                            // the root.
    BinaryNode *right = r->right;
                            //Now make sure that the root of the right tree
    pushRevDown(right);     //has an up-to-date index:
    pushIndDown(right);

    r->right = nullNode;    //Remove the right subtree:
    right->parent = nullNode;
    updateFields(r);

    return right;
}

    //assume: all indices of root < all indices of root2
    //root modified, root2 is made right subtree of root
    //changed root returned
Permutation::BinaryNode* Permutation::join(BinaryNode *&r, BinaryNode *root2)
{
    if(root2 != nullNode){
        splay(r->size-1, r);

        pushRevDown(r);
        pushIndDown(r);
        r->right = root2;
        root2->parent = r;
        updateFields(r);
    }
    return r;
}

ostream& operator<<(ostream& output, Permutation& p)
  {
  output << p.getSeqVec();
  return output;
  }

