#ifndef CYCLE_H
#define CYCLE_H

/*
A Cycle contains a list of nodes.
It can be good or bad.  Since a cycle of only 2 nodes cannot be bad we
  assume it is GOOD until a third node is added at which point we assume it
  is BAD.
*/


#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <vector>

using namespace std;

class Cycle;
std::ostream& operator<<(std::ostream& output, const Cycle& c);

class Cycle
  {
  friend std::ostream& operator<<(std::ostream& output, const Cycle& c);

  public:
    Cycle();
    Cycle(const Cycle& c);
    ~Cycle();
    void operator=(const Cycle& c);

    bool getGood() const;       //Return true if this is a good cycle.
    void setGood(bool g);
        //Return true if this has an associated deletion.
        //   Reset the cycle's value.  ** Temporary
    bool getDel();
        //Say that this cycle has an associated deletion.  ** Temporary
    void setDel();
    int getSize() const;
    void print() const;
    void printNodes(ostream& output) const;
    void clear();
    void add(int node, int permnum);   //Add a node to the cycle and set its
                                       //identity for the permutation.
    vector<int>::iterator getBeg();    //Get the beginning of the node list.
    vector<int>::iterator getEnd();    //Get the end of the node list.
    int getPermNum(int i);             //Get the ith member of -permnums-.
    int getNodeNum(int i);             //Get the ith member of -nodes-.

  private:
    void copy(const Cycle& c);

    bool good;            //true if this is a good(oriented) cycle.
    bool del;             //true if this has a deletion associated with it.
    vector<int> nodes;    //The list of node indices.
    vector<int> permnums; //Permutation element that the node corresponds to.
  };


#endif
