/* Krister Swenson                                                 Winter 2007-8

Sort the given permutation my repeatedly applying a maximum set of noninterfering inversions.
*/

#include <iostream>
#include <vector>
#include <algorithm>
#include <numeric>
#include <math.h>
#include "../graph/Ograph.h"
#include "../tools/tools.h"


using namespace std; 



bool isCommuting(const list<Pair> &set)
  {
  for(list<Pair>::const_iterator i = set.begin(); i != set.end(); i++)
    {
    list<Pair>::const_iterator j = i;
    while(++j != set.end())
      if(i->overlaps(*j))     //If the two intervals overlap then sort the 
        {                     //intervals and print message:
        vector<Pair> setv;
        std::copy(set.begin (), set.end (), back_inserter(setv));
        sort(setv.begin(), setv.end());
        cerr << "sorted NMIS:" << setv << endl;
        cerr << "offending pairs:" << *i << *j << endl;

        return false;
        }
    }

  return true;
  }



/* Run a test for a single permutation.
*/
void tryPerm(Ograph g)
  {
  int numreps = 0;
  long long tracesize = 1;
  unsigned long long nodesum = 0;
  int dist = g.getDistNH();
  while(dist)                             //While it's not sorted
    {
    numreps++;
    list<Pair> nmis = g.getNonintMIS();   //find a maximum noninterfering set,
    long long size = factorial(nmis.size());
    unsigned long long numnodes = exp2(nmis.size());
    cerr << "\tNMIS(" << nmis.size() << "!=" << size << " paths and "
         << numnodes << " nodes):" << endl << nmis << endl;
    tracesize *= size;
    nodesum += numnodes;
                                          //Sanity check:  ___________________
    assert(g.isNonint(nmis));             // Make sure they sort (and commute).
    assert(isCommuting(nmis));            // Double check they commute.
  
    g.applyCommutingSet(nmis);            //then apply the set,
    dist = g.getDistNH();                 //and check the new distance.
    }

  cerr << "depth:" << numreps << endl;
  cerr << "trace size >= " << tracesize << endl;
  cerr << "number of nodes >= " << nodesum << endl;
  }




int main()
  {
  cout << "seq:";
  vector<int> perm = getInputSeq();
  Ograph g(perm);
  tryPerm(g);

  return(0);
  }
