// This is the Bader, Moret, Yan linear-time inversion distance.
// I added a wrapper so you can just give it a permutation.
//
#ifndef INVDIST_H
#define INVDIST_H

#include <iostream>
#include <vector>
#include "seqfuncs.h"

#ifndef CYGWINNT
#include <sys/time.h>
#else
#include <time.h>
#endif

using namespace std;

extern double time_linear;

struct genome_struct
{
    int *genes;
    int genome_num;
    char *encoding;
    char *gnamePtr;             /* Used to copy the gname when adding genomes to the tree */
    char parent[32];
};

typedef struct
{
    int index;                  /* index of component's root */
    int oriented;               /* Boolean: Is component oriented */
    int blocks;                 /* Number of blocks in nonoriented component */
    int hurdle;                 /* Bitmask of HURDLE properties.
                                   Bit 0 = hurdle
                                   Bit 1 = wrap-around hurdle
                                   Bit 2 = superhurdle */
    int left;                   /* Index of component to the left of my rightmost block */
    int right;                  /* Index of component to the right of my rightmost block */
} component_t;

#define HURDLE          1
#define GREATHURDLE (1<<1)
#define SUPERHURDLE (1<<2)

typedef struct
{
    int *hammingArr;            /* DIST_BP:  (NUMGENES+1)*2 */

    int *perm1;                 /* DIST_INV: 2*num_genes + 2 */
    int *perm2;
    int *perm;
    int *done;
    int *greyEdges;

    int *stack;                 /* DIST_INV: size */
    int *oriented;
    int *cc;
    int *labeled;
    component_t *components;

} distmem_t;


// Return the reversal distance from the given permutation to the identity.
int getLinearTimeDist(vector<int> perm);
int getLinearTimeDist(int* perm, int len);
// Return a vector where the ith entry is true if the ith reality edge is
// oriented.
vector<bool> getRealityEdgeOrientations(int* perm, int len);
vector<bool> getRealityEdgeOrientations(vector<int> perm);

void calc_invmatrix ( struct genome_struct *genomes, int num_genes,
                      int num_genomes, distmem_t * distmem, int CIRCULAR );

void setinvmatrix ( int **distmatrix, struct genome_struct *genomes,
                    int num_genes, int num_genomes, distmem_t * distmem,
                    int CIRCULAR );

int invdist_noncircular ( struct genome_struct *g1, struct genome_struct *g2,
                          int offset, int num_genes, distmem_t * distmem );
int invdist_circular ( struct genome_struct *g1, struct genome_struct *g2,
                       int num_genes, distmem_t * distmem );


int invdist_noncircular_nomem ( struct genome_struct *g1,
                                struct genome_struct *g2,
                                int offset, int num_genes );
int invdist_circular_nomem ( struct genome_struct *g1,
                             struct genome_struct *g2, int num_genes );

int calculate_offset ( struct genome_struct *g1, struct genome_struct *g2,
                       int num_genes );

void connected_component ( int size, distmem_t * distmem,
                           int *num_components );

/* Creates and returns a distmem_t structure for use in calculating
   pairwise inversion distances */
distmem_t *new_distmem ( int NUM_GENES );
/* Frees all memory associated with the specified distmem_t structure
   (including that allocated for the structure itself) */
void free_distmem ( distmem_t * distmem );

#endif
