/* Krister Swenson                                                  Spring 2006

Function used on sequences.
*/

#include <ext/hash_map>
#include "seqfuncs.h"
#include "../tools/tools.h"
#include <stdlib.h>

using namespace std; 
using namespace __gnu_cxx; 




/*  Return true if there is a hurdle in the given permutation.
*/
bool hasHurdle(const vector<int> p)
  { return getUnorientedComps(p).size(); }


/*
Input: Permutation
Output: Indices of endpoints of each hurdle in the permutation or empty vector if there is no hurdle
*/
typedef enum {PLUS, MINUS} sign;
vector<ipair> getUnorientedComps(const vector<int> perm)
{
  vector<int> p(perm.size());
  p.push_back(0);
  copy(perm.begin(), perm.end(), p.begin()+1);
  p.push_back(perm.size()+1);

  stack<int> s1;
  stack<int> s2;
  stack<int> m1;
  stack<int> m2;

  vector<ipair> hurdleList;

  int n = p.size();
  int *max = new int[n];
  int *min = new int[n];
  int *M = new int[n];
  int *m = new int[n];
  bool *mark1 = new bool[n];
  bool *mark2 = new bool[n];
  int *pi = new int[n];
  sign *si = new sign[n];

  for(int i=0; i < n; i++)
  {
    if(p.at(i) > 0)
    {
      pi[i] = p.at(i);
      si[i] = PLUS;
    }
    else
    {
      pi[i] = 0 - p.at(i);
      si[i] = MINUS;
    }
  }

  M[0] = n-1;
  m[0] = 0;
  max[0] = 0;
  min[0] = n-1;
  mark1[0] = false;
  mark2[0] = false;
  s1.push(0);
  s2.push(0);
  m1.push(n);
  m2.push(0);

  int t;

  for(int i = 1; i < n; i++)
  {

    //compute M[i]
    if (pi[i-1] > pi[i])
    {
      M[i] = pi[i-1];
      m1.push(pi[i-1]);
    }
    else
    {
      while(m1.top() < pi[i])
        m1.pop();
      M[i] = m1.top();
    }
    //find connected components
    while(pi[i] < pi[(t = s1.top())] || pi[i] > M[t])
    {
      s1.pop();
      max[s1.top()] = (max[s1.top()] > max[t])? max[s1.top()] : max[t];
      mark1[s1.top()] |= mark1[t];
    }
    if((si[i] == PLUS) && (pi[i] == max[t = s1.top()]+1) &&
       (i-t == pi[i] - pi[t]) && (i-t > 1))
    {
      if(mark1[t] == false)
      {
        ipair hurdle;
        hurdle.first = t;
        hurdle.second = i;
        assert(t < i);
        hurdleList.push_back(hurdle);
      }
      mark1[s1.top()] = false;
    }

    //compute m[i]
    if (pi[i-1] < pi[i])
    {
      m[i] = pi[i-1];
      m2.push(pi[i-1]);
    }
    else
    {
      while(m2.top() > pi[i])
        m2.pop();
      m[i] = m2.top();
    }
    //find connected components
    while((pi[i] > pi[(t = s2.top())] || pi[i] < m[t]) && t>0)
    {
      s2.pop();
      min[s2.top()] = (min[s2.top()] < min[t])? min[s2.top()] : min[t];
      mark2[s2.top()] |= mark2[t];
    }
    if((si[i] == MINUS) && (pi[i] == min[t = s2.top()]-1) &&
       (i-t == pi[t] - pi[i]) && (i-t > 1))
    {
      if(mark2[t] == false)
      {
        ipair hurdle;
        hurdle.first = t;
        hurdle.second = i;
        assert(t < i);
        hurdleList.push_back(hurdle);
      }
      mark2[s2.top()] = false;
    }

    if (si[i] == PLUS)
      s1.push(i);
    else
      s2.push(i);
    max[i] = pi[i];
    min[i] = pi[i];
    min[s2.top()] = (min[s2.top()] < pi[i]) ? min[s2.top()] : pi[i];
    max[s1.top()] = (max[s1.top()] > pi[i]) ? max[s1.top()] : pi[i];
    mark1[s1.top()] = si[i]==MINUS;
    mark2[s2.top()] = si[i]==PLUS;
  }

  delete []max;
  delete []min;
  delete []M;
  delete []m;
  delete []mark1;
  delete []mark2;
  delete []pi;
  delete []si;

  return hurdleList;
}




/*Return a count of all permutations of length -len- with len+1 breakpoints.
   ** much less resource dependent than getBPPerm.
*/
unsigned long getBPPermCount(int len)
  {
  unsigned long oneback, twoback, threeback;
  threeback = getBPPerm(1).size();
  twoback = getBPPerm(2).size();
  oneback = getBPPerm(3).size();
  if(len < 0)
    return 0;
  if(len == 1)
    return threeback;
  if(len == 2)
    return twoback;

  int i = 3;
  unsigned long Qi = oneback;
  while(i != len)
    {
    i++;
    Qi = 2*(i-1)*oneback + (4*i-3)*twoback + (2*i-2)*threeback;
    threeback = twoback;
    twoback = oneback;
    oneback = Qi;
    }

  return Qi;
  }

/*Return a count of all positive permutations of length -len- with len+1
breakpoints.
*/
unsigned long getBPPosCount(int len)
  {
  unsigned long oneback, twoback, threeback;
  threeback = 0;
  twoback = 1;
  oneback = 1;
  if(len < 0)
    return 0;
  if(len == 1)
    return threeback;
  if(len == 2)
    return twoback;

  int i = 3;
  unsigned long Qi = oneback;
  while(i != len)
    {
    i++;
    Qi = (i-1)*oneback + (2*i-3)*twoback + (i-1)*threeback;
    threeback = twoback;
    twoback = oneback;
    oneback = Qi;
    }

  return Qi;
  }

/*Return a vector of all permutations of length -len- with len+1 breakpoints.
*/
vector< vector<int> > getBPPerm(int len)
  {
  vector<int> s;
  vector< vector<int> > ret;
  getBPPerm(ret, s, makeIDv(len), len);
  return ret;
  }
void getBPPerm(vector< vector<int> > &ret, vector<int> seq, vector<int> L,
               int SEQLEN)
  {
  if(L.empty())            //If we're out of characters to place then
    ret.push_back(seq);    // add the full permutation.

  for(vector<int>::iterator i = L.begin(); i != L.end(); i++)
    {
    int temp = *i;
    vector<int>::iterator j = L.erase(i);

    if(seq.empty() || (temp-*seq.rbegin() != 1)) //If it's a breakpoint
      {
      if(!(seq.empty() && temp == 1) && !(L.empty() && temp == SEQLEN))
        {
        seq.push_back(temp);                     //add it to the sequence
        getBPPerm(ret, seq, L, SEQLEN);          //and recurse.
        seq.pop_back();
        }
      }
    if(seq.empty() || (-temp-*seq.rbegin() != 1))//If it's a breakpoint
      {
      temp *= -1;
      seq.push_back(temp);                       //add it to the sequence
      getBPPerm(ret, seq, L, SEQLEN);            //and recurse.
      seq.pop_back();
      }

    L.insert(j, temp);
    }
  }



/*Return true if the given permutation has an adjacency.
*/
bool hasAdjacency(const vector<int> p)
  {
  if(p.front() == 1 || p.back() == (int)p.size())
    return true;
  for(vector<int>::const_iterator i=p.begin(); i!=p.end();)
    {
    vector<int>::const_iterator j=i;
    i++;
    if(*i - *j == 1)
      return true;
    }
  return false;
  }

/*Apply the given reversal to the given permutation.
*/
void applyRev(vector<int> &p, ipair rev)
  {
  for(int i=0; i < (rev.second-rev.first+2)/2; i++)
    {
    int tmp = p[rev.first-1+i];
    p[rev.first-1+i] = -p[rev.second-1-i];
    p[rev.second-1-i] = -tmp;
    }
  }

/*Return a vector version of the given permutation.
*/
vector<int> getVector(int* perm, int len)
  {
  vector<int> p;
  p.resize(len);
  for(int i=0; i<len; i++)
    p[i] = perm[i];

  return p;
  }

/*Return a random permutation.
*/
vector<int> getRandPerm(int len)
  {
  vector<int> perm;

  for(int i=1; i <= len; i++)        //Create the identity:
    perm.push_back(i);

  for(unsigned int j=0; j < perm.size(); j++)
    {                                //Randomize the permutation:
    int k = random()%perm.size();    //Swap perm[j] with a random element.
    int temp = perm[k];
    perm[k] = perm[j];
    perm[j] = temp;
    }

  return perm;
  }

/*Return a random signed permutation.
*/
vector<int> getRandSPerm(int len)
  {
  vector<int> perm = getRandPerm(len);

  for(unsigned int j=0; j < perm.size(); j++)
    if(random()%2)                   //Assign signs randomly:
      perm[j] *= -1;

  return perm;
  }


/*Read in an input sequence and return an integer array that is the sequence or
return 0 if the end of file is reached.
  ** User resposible for the memory.
*/
int getInputSeq(int** ret)
  {
  int tmp[MAXSEQLEN];

  char* in = NULL;
  size_t size = getline(&in, &size, stdin);
  if((int)size <= 1)
    return 0;

  char* inindex = in;
  int i=0;

  int num;
  int len;
  sscanf(inindex, "%i %n", &num, &len);
  inindex += len;
  size -= len;
  tmp[i++] = num;
  while(size)
    {
    sscanf(inindex, "%i %n", &num, &len);
    tmp[i++] = num;
    inindex += len*sizeof(char);
    size -= len;
    assert(i < MAXSEQLEN);
    }
  free(in);

  int newsize = i;
  *ret = (int*)calloc(newsize, sizeof(int));
  while(--i >= 0)
    (*ret)[i] = tmp[i];

  return newsize;
  }

vector<int> getInputSeq()
  {
  int *seq=0;
  int len = getInputSeq(&seq);
  vector<int> retval(seq, seq + len);
  free(seq);
  return retval;
  }


/* Label seq2 as though we were renaming s1 to be the identity.
*/
void seqACCseq(const int s1[], int s2[], int len)
  {
  int map[len+1];             //We won't use the 0 index (simpler code).
  for(int i=0; i < len; i++)
    {
    if(s1[i] > 0)
      map[s1[i]] = i+1;
    else
      map[abs(s1[i])] = -(i+1);
    }

  for(int i=0; i < len; i++)
    {
    if(s2[i] > 0)
      s2[i] = map[s2[i]];
    else
      s2[i] = -map[abs(s2[i])];
    }
  }


/* Label seq2 as though we were renaming s1 to be the identity.
*/
void seqACCseq(const vector<int> s1, vector<int> &s2)
  {
  int map[s1.size()+1];           //We won't use the 0 index (simpler code).
  for(unsigned i=0; i < s1.size(); i++)
    {
    if(s1[i] > 0)
      map[s1[i]] = i+1;
    else
      map[abs(s1[i])] = -(i+1);
    }

  for(unsigned i=0; i < s1.size(); i++)
    {
    if(s2[i] > 0)
      s2[i] = map[s2[i]];
    else
      s2[i] = -map[abs(s2[i])];
    }
  }



/* Make sure that the given sequence is a valid permutation of the identity.
*/
bool validPerm(const int seq[], const int len)
  {
  hash_map< int, int, hash<int> > beenseen; //True if the value has been seen.
  for(int i=0; i < len; i++)         //Check that a value is not there more
    {                                //than once.
    if(beenseen[abs(seq[i])] > 1)
      return false;
    else
      beenseen[abs(seq[i])]++;
    }
  for(int i=1; i <= len; i++)        //Now check that every value is there.
    {
    if(beenseen[i] != 1)
      return false;
    }

  return true;
  }


/* Rename the given sequence to be the identity; return the identity renamed
accordingly.
*/
void invertSeq(vector<int> seq)
  {
  vector<int> ts;
  ts.resize(seq.size());

  for(unsigned i=0; i<seq.size(); i++)
    {
    if(seq[i] > 0)
      ts[abs(seq[i])-1] = i+1;
    else
      ts[abs(seq[i])-1] = -(i+1);
    }


  for(unsigned i=0; i<seq.size(); i++)
    seq[i] = ts[i];
  }

void invertSeq(int seq[], int len)
  {
  int ts[len];

  for(int i=0; i<len; i++)
    {
    if(seq[i] > 0)
      ts[abs(seq[i])-1] = i+1;
    else
      ts[abs(seq[i])-1] = -(i+1);
    }


  for(int i=0; i<len; i++)
    seq[i] = ts[i];
  }

/* Print a sequence.
*/
void printSeq(const int s[], const int slen)
  {
  for(int i=0; i<slen; i++)
    cerr << s[i] << " ";
  }


/* Return s1 renamed as though s2 were the target.
*/
int* renameSeq(const int s1[], const int s1len, const int s2[], const int s2len)
  {
  int* ret = (int*)calloc(s1len, sizeof(int));

  assert(s1len == s2len);    //Ensure the sequence lengths match.

  int index[s2len];          //index[i] is the index where i is located in s2.
                             // (so should be the value that i becomes in s1)
  for(int i=0; i<s2len; i++)     //Build the index:
    {
    if(s2[i] < 0)                //If the number is negative
      index[abs(s2[i])] = -(i+1);//then store its index as negative
    else                         //otherwise,
      index[s2[i]] = (i+1);      //store it as positive.
    }

  for(int j=0; j<s1len; j++)     //Use the index.
    {
    if(s1[j]>0)
      ret[j] = index[abs(s1[j])];
    else
      ret[j] = -index[abs(s1[j])];
    }

  return ret;
  }

/*Return the identity of the given length.
  ** User resposible for the memory.
*/
int* makeID(const int len)
  {
  int* ret = (int*)calloc(len, sizeof(int));
  for(int i=1; i <= len; i++)
    ret[i-1] = i;
  return ret;
  }
vector<int> makeIDv(const int len)
  {
  vector<int> ret;
  for(int i=1; i <= len; i++)
    ret.push_back(i);
  return ret;
  }



/*Copy the sequence.
  ** User resposible for the memory.
*/
int* copySeq(const int* s, const int len)
  {
  int* ret = (int*)calloc(len, sizeof(int));
  for(int i=0; i < len; i++)
    ret[i] = s[i];
  return ret;
  }


/* Return a string from the given input.
*/
template < class T >
string ToString(const T &arg)
  {
  ostringstream out;
  out << arg;

  return(out.str());
  }

/* Return a string of the given sequence.
*/
string seqTOstring(int* seq, int seqsize)
  {
  string ret = ToString(seq[0]);

  for(int i=1; i<seqsize; i++)
    ret += " " + ToString(seq[i]);

  return ret;
  }

