/* Krister Swenson                                                  Spring 2006

Function used for finding signatures.
 **NEED TO MODIFY:  the multiple signature function take the inverse of the
                    graphs needed.
*/

#include "sigfuncs.h"
#include <ext/hash_map>


using namespace std; 
using namespace __gnu_cxx;




/*Return all maximum signatures between -g1- and -g2-.
*/
vector<Ograph> getSigs(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getSigs_R(gc1,gc2,idg,0);
  }
vector<Ograph> getSigs_R(Ograph &g1, Ograph &g2, const Ograph &idg,
                         const int depth)
  {
  vector<Pair> oOps1, oOps2;
  int maxdepth = depth;
  vector<Ograph> maxgs;
  maxgs.push_back(idg);

  if(g1.notSorted() && g2.notSorted())
    {
    g1.getSafeInvs(oOps1);       //Get the operations on divergent edges.
    g2.getSafeInvs(oOps2);       //Get the operations on divergent edges.

    for(int i=0; size_t(i) < oOps1.size(); i++)
      {                              //For each divergent op in s1
      for(int j=0; size_t(j) < oOps2.size(); j++)
        {
        if(oOps1[i] == oOps2[j])     //that matches one in s2
          {
          Ograph newg1 = g1;         //make a new graph
          newg1.revAll(oOps1[i]);    //that we do the inversion on.
          Ograph newg2 = g2;         //make a new graph
          newg2.revAll(oOps2[j]);    //that we do the inversion on.
          Ograph newidg = idg;       //make a new graph
          newidg.revAll(oOps2[j]);   //that we do the inversion on.

          Ograph g = getSig_R(newg1, newg2, newidg, depth+1);
          if(g.getDistNH() > maxdepth)
            {
            maxdepth = g.getDistNH();
            maxgs.clear();
            maxgs.push_back(g);
            }
          else if((g.getDistNH() == maxdepth) &&
                  (find(maxgs.begin(), maxgs.end(), g) == maxgs.end()))
            maxgs.push_back(g);
          }
        }
      }
    }

  return maxgs;          //Return the maximum depth graphs that we found.
  }



/*Return a largest signature between -g1- and -g2-.
*/
Ograph getSig(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getSig_R(gc1,gc2,idg,0);
  }
Ograph getSig_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth)
  {
  vector<Pair> oOps1, oOps2;
  int maxdepth = depth;
  Ograph maxg = idg;

  if(g1.notSorted() && g2.notSorted())
    {
    g1.getSafeInvs(oOps1);       //Get the operations on divergent edges.
    g2.getSafeInvs(oOps2);       //Get the operations on divergent edges.

    for(int i=0; size_t(i) < oOps1.size(); i++)
      {                              //For each divergent op in s1
      for(int j=0; size_t(j) < oOps2.size(); j++)
        {
        if(oOps1[i] == oOps2[j])     //that matches one in s2
          {

          Ograph newg1 = g1;         //make a new graph
          newg1.revAll(oOps1[i]);    //that we do the inversion on.
          Ograph newg2 = g2;         //make a new graph
          newg2.revAll(oOps2[j]);    //that we do the inversion on.
          Ograph newidg = idg;       //make a new graph
          newidg.revAll(oOps2[j]);   //that we do the inversion on.

          Ograph g = getSig_R(newg1, newg2, newidg, depth+1);
          if(g.getDistNH() > maxdepth)
            {
            maxdepth = g.getDistNH();
            maxg = g;
            }
          }
        }
      }
    }

  return maxg;           //Return the maximum depth graph that we found.
  }




/*Return the size of the largest signature between -g1- and -g2-.
*/
int getSigSize(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getSigSize_R(gc1,gc2,idg,0);
  }
int getSigSize_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth)
  {
  vector<Pair> oOps1, oOps2;
  int maxdepth = depth;

  if(g1.notSorted() && g2.notSorted())
    {
    g1.getSafeInvs(oOps1);       //get the operations on divergent edges,
    g2.getSafeInvs(oOps2);       //get the operations on divergent edges,

    for(int i=0; size_t(i)<oOps1.size(); i++)
      {                              //for each divergent op
      for(int j=0; size_t(j)<oOps2.size(); j++)
        {
        if(oOps1[i] == oOps2[j])     //that matches one in s2
          {
          Ograph newg1 = g1;         //make a new graph
          newg1.revAll(oOps1[i]);    //that we do the inversion on.
          Ograph newg2 = g2;         //make a new graph
          newg2.revAll(oOps2[j]);    //that we do the inversion on.
          Ograph newidg = idg;       //make a new graph
          newidg.revAll(oOps2[j]);   //that we do the inversion on.

          int d = getSigSize_R(newg1, newg2, newidg, depth+1);
          if(d > maxdepth)
            maxdepth = d;
          }
        }
      }
    }

  return maxdepth;           //Return the maximum depth that we found.
  }


/* Return the depth of the largest signature found via the greedy method.
*/
int getGreedySig(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getGreedySig_R(gc1,gc2,idg,0);
  }
int getGreedySig_R(Ograph &g1, Ograph &g2, const Ograph &idg, int depth)
  {
  vector<Pair> oComOps;
  int largest=depth;

  if(g1.notSorted() && g2.notSorted())
    {
    getBestCommon(g1, g2, oComOps);

cerr << "TIES:" << oComOps.size() << " [" << depth << "]" << endl;
if(oComOps.size() == 0)
  {
  cerr << "-----TOP:" << depth << endl;
  //idg.printSeqStr(); cerr << endl;
  }
    for(int i=0; size_t(i)<oComOps.size(); i++)
      {                            //for each best common sorting op
if(i>0)
  {
  if(oComOps[i] == oComOps[i-1])
    {
    cerr << "\n\nOUTTA" << endl;
    exit(0);
    }
  }
      Ograph newg1 = g1;           //make a new graph
      newg1.revAll(oComOps[i]);    //that we do the inversion on.
      Ograph newg2 = g2;           //make a new graph
      newg2.revAll(oComOps[i]);    //that we do the inversion on.
      Ograph newidg = idg;         //make a new graph
      newidg.revAll(oComOps[i]);   //that we do the inversion on.

      int d = getGreedySig_R(newg1, newg2, newidg, depth+1);//continue to sort.
      if(largest < d)
        largest = d;
      }
    }

  return largest;
  }



/* Return the depth of the largest signature found via the greedy method.
This version is faster because it keeps track of the sequences already visited.
*/
//  A global hash_set that keeps all sequences we've come across so far:
hash_set<const char*, hash<const char*>, eqstr> beenseen;
int getGreedySigF(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getGreedySigF_R(gc1,gc2,idg,0);
  }
int getGreedySigF_R(Ograph &g1, Ograph &g2, const Ograph &idg, int depth)
  {
  vector<Pair> oComOps;
  int largest=depth;

  char* ids = idg.getSeqStr();
  beenseen.insert(ids);
  free(ids);

  if(g1.notSorted() && g2.notSorted())
    {
    getBestCommon(g1, g2, oComOps);

cerr << "TIES:" << oComOps.size() << " [" << depth << "]" << endl;
if(oComOps.size() == 0)
  {
  cerr << "-----TOP:" << depth << endl;
  ////idg.printSeqStr(); cerr << endl;
  }

    for(int i=0; size_t(i)<oComOps.size(); i++)
      {                            //For each best common sorting op
      Ograph newidg = idg;         //make a new graph (from the id)
      newidg.revAll(oComOps[i]);   //that we do the inversion on.
      char* newids = newidg.getSeqStr();
      if(!lookup(beenseen, newids))  //If we haven't gone up that path then
        {
        Ograph newg1 = g1;           //make a new graph
        newg1.revAll(oComOps[i]);    //that we do the inversion on,
        Ograph newg2 = g2;           //make a new graph
        newg2.revAll(oComOps[i]);    //that we do the inversion on,
                                     //and continue to sort.
        int d = getGreedySigF_R(newg1, newg2, newidg, depth+1);
        if(largest < d)
          largest = d;
        }
      else
        cerr << "BACK DOWN" << endl;
      free(newids);
      }
    }

  return largest;
  }

/*Lookup function for the beenseen hash_set.
*/
bool lookup(const hash_set<const char*, hash<const char*>, eqstr>& Set,
            const char* word)
{
  hash_set<const char*, hash<const char*>, eqstr>::const_iterator it
    = Set.find(word);
  return (it != Set.end());
}



/* Return the depth of the largest signature found via the greedy method where
the first tie is always chosen.
*/
int getGreedySigSP(Ograph &g1, Ograph &g2, const Ograph &idg)
  {
  Ograph gc1 = g1.getInvGraph();
  Ograph gc2 = g2.getInvGraph();
  return getGreedySigSP_R(gc1,gc2,idg,0);
  }
int getGreedySigSP_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth)
  {
  vector<Pair> oComOps;
  int largest=depth;

  if(g1.notSorted() && g2.notSorted())
    {
    getBestCommon(g1, g2, oComOps);

    cerr << oComOps.size() << "[" << depth << "] ";
    if(oComOps.size() != 0)
      {
      Ograph newg1 = g1;           //make a new graph
      newg1.revAll(oComOps[0]);    //that we do the inversion on,
      Ograph newg2 = g2;           //make a new graph
      newg2.revAll(oComOps[0]);    //that we do the inversion on,
      Ograph newidg = idg;         //make a new graph
      newidg.revAll(oComOps[0]);   //that we do the inversion on,
                                   //and continue deeper.
      largest = getGreedySigSP_R(newg1, newg2, newidg, depth+1);
      }
    else
      cout << endl << idg.getSeqStr() << endl;
    }
  else
    {
    cerr << "sorted.";
    cout << endl << idg.getSeqStr() << endl;
    }

  return largest;
  }
/*
int getGreedySigSP(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth)
  {
  int largest=depth;

  if(g1.notSorted() && g2.notSorted())
    {
    Pair inv = getSingleBestCommon(g1, g2);

    cerr << "[" << depth << "] ";
    if(!(inv == Pair(0,0)))
      {
      Ograph newg1 = g1;           //make a new graph
      newg1.revAll(inv);           //that we do the inversion on,
      Ograph newg2 = g2;           //make a new graph
      newg2.revAll(inv);           //that we do the inversion on,
      Ograph newidg = idg;         //make a new graph
      newidg.revAll(inv);          //that we do the inversion on,
                                   //and continue deeper.
      largest = getGreedySigSP(newg1, newg2, newidg, depth+1);
      }
    else
      cout << endl << idg.getSeqStr() << endl;
    }
  else
    {
    cerr << "sorted.";
    cout << endl << idg.getSeqStr() << endl;
    }

  return largest;
  }
*/



/* Return the depth of the largest signature found for the given set of
sequences via the greedy method where the first tie is always chosen.
  ** Takes a vector of INVERSE graphs.
*/
int getGreedySigSP(vector<Ograph> &graphs)
  {                   //For convenience, ensure the closest is the first:
  int min=graphs[0].getDist();
  int minind=0;
  for(int i=1; size_t(i)<graphs.size(); i++)
    {
    int dist = graphs[i].getDist();
    if(min > dist)
      {
      min = dist;
      minind = i;
      }
    }
  Ograph tmp = graphs[0];
  graphs[0] = graphs[minind];
  graphs[minind] = tmp;

  int len = graphs[0].getSeqLen();
  Ograph id(makeID(len), len);
  return getGreedySigSP(graphs, id);
  }
int getGreedySigSP(vector<Ograph> &graphs, const Ograph &idg, const int depth)
  {
  vector<Pair> oComOps;
  int largest=depth;

  if(graphs[0].notSorted())
    {
    getBestCommon(graphs, oComOps);

    cerr << oComOps.size() << "[" << depth << "] ";
    if(oComOps.size() != 0)
      {
      vector<Ograph> newgraphs = graphs;
      for(int i=0; size_t(i)<newgraphs.size(); i++)
        newgraphs[i].revAll(oComOps[0]);

      Ograph newidg = idg;         //make a new graph
      newidg.revAll(oComOps[0]);   //that we do the inversion on,
                                   //and continue deeper.
      largest = getGreedySigSP(newgraphs, newidg, depth+1);
      }
    else
      cout << endl << idg.getSeqStr() << endl;
    }
  else
    {
    cerr << "sorted.";
    cout << endl << idg.getSeqStr() << endl;
    }

  return largest;
  }




/*Print every node of the lattice that is a signature.
For searching for paths of -depth- up the lattice the value is the
depth to truncate the upwards push at; if negative we just print the lattice.
Immediately stop search and return true if the upwards push was truncated.
*/
#define NOPRINT 0   //If 1 we don't print links when depth is non-negative.
bool printSigLattice(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth)
  {
  vector<Pair> oOps1, oOps2;
  char hurdlewarning[] = "POSITIVE PERMUTATION!";

  if(g1.notSorted() && g2.notSorted())
    {
    g1.getSafeInvs(oOps1);       //get the operations on divergent edges,
    g2.getSafeInvs(oOps2);       //get the operations on divergent edges,

    for(int i=0; size_t(i)<oOps1.size(); i++)
      {                              //for each divergent op
      char *g1str = g1.getSeqStr();
      char *idgstr = idg.getSeqStr();
      char *pointer1 = idgstr;
    
      for(int j=0; size_t(j)<oOps2.size(); j++)
        {
        if(oOps1[i] == oOps2[j])     //that matches one in s2
          {
          Ograph newg1 = g1;         //make a new graph
          newg1.revAll(oOps1[i]);    //that we do the inversion on.
          Ograph newg2 = g2;         //make a new graph
          newg2.revAll(oOps2[j]);    //that we do the inversion on.
          Ograph newidg = idg;       //make a new graph
          newidg.revAll(oOps2[j]);   //that we do the inversion on.

          if(depth < 0 || !NOPRINT)  //If we're not searching for a path 
            {                        //greater than -depth- we print the links.
            char *g2str = g2.getSeqStr();
            char *ng1str = newg1.getSeqStr();
            char *ng2str = newg2.getSeqStr();
            char *nidgstr = newidg.getSeqStr();
            char *pointer2 = nidgstr;
            if(g1.isPositive() || g2.isPositive())
              idgstr = hurdlewarning;
            if(newg1.isPositive() || newg2.isPositive())
              nidgstr = hurdlewarning;
            outputLink(g1str, g2str, idgstr, idg.getDistNH(),
                       ng1str, ng2str, nidgstr, newidg.getDistNH(), depth);
            free(g2str); free(ng1str); free(ng2str); free(pointer2);
            }

          if(depth != 0)
            {
            if(printSigLattice(newg1, newg2, newidg, depth-1))
              return true;
            }
          else
            return true;  //Return the fact that we've truncated.
          }
        }
      free(g1str);
      free(pointer1);
      }
    }

  return false;           //Return that fact that we haven't truncated.
  }


/*Print every node of the signature path that maximizes the number of shared 
common sorting inversion in the resulting permutation.
-depth- should always be given as 0.

Return the depth of the largest signature found (plus 1).
*/
int printGreedy(Ograph &g1, Ograph &g2, const Ograph &idg, int depth)
  {
  vector<Pair> oComOps;
  int largest=depth;
  char hurdlewarning[] = "POSITIVE PERMUTATION!";

  if(g1.notSorted() && g2.notSorted())
    {
    int indegree = getBestCommon(g1, g2, oComOps);

    for(int i=0; size_t(i)<oComOps.size(); i++)
      {                                //for each best common sorting op
      Ograph newg1 = g1;           //make a new graph
      newg1.revAll(oComOps[i]);    //that we do the inversion on.
      Ograph newg2 = g2;           //make a new graph
      newg2.revAll(oComOps[i]);    //that we do the inversion on.
      Ograph newidg = idg;         //make a new graph
      newidg.revAll(oComOps[i]);   //that we do the inversion on.

      char *g1str = g1.getSeqStr();
      char *g2str = g2.getSeqStr();
      char *idgstr = idg.getSeqStr();
      char *pointer1 = idgstr;
      char *ng1str = newg1.getSeqStr();
      char *ng2str = newg2.getSeqStr();
      char *nidgstr = newidg.getSeqStr();
      char *pointer2 = nidgstr;
      if(g1.isPositive() || g2.isPositive())
        idgstr = hurdlewarning;
      if(newg1.isPositive() || newg2.isPositive())
        nidgstr = hurdlewarning;
      cerr << "in-degree[" << depth << "]:" << indegree << endl;
      outputLink(g1str, g2str, idgstr, idg.getDistNH(),
                 ng1str, ng2str, nidgstr, newidg.getDistNH(), -1);

      int d = printGreedy(newg1, newg2, newidg, depth+1); //and continue to sort.
      if(largest < d)
        largest = d;

      free(g1str);
      free(g2str);
      free(ng1str);
      free(ng2str);
      free(pointer2);
      free(pointer1);
      }
    }

  return largest;
  }


/*Print the dead-end path down as far as possible towards the inverse identity.
In other word, unsort the sequences so that they don't gain some path that can
move them upwards through the lattice and past the starting spot.

The -depth- should be the upwards distance past the given sequence the upwards
search should stop at, normally 1.
*/
void printDeadend(Ograph &g1, Ograph &g2, Ograph &idg, const int depth)
  {
  vector<Pair> uOps1, uOps2, oOps;

  g1.getUnsortingInvs(uOps1); //Get the operations that unsort from the bottom.
  g2.getUnsortingInvs(uOps2); //Get the operations that unsort from the bottom.
  idg.getSafeInvs(oOps);      //Get the operations that sort to the id.

  for(int i=0; size_t(i)<uOps1.size(); i++)
    {
    for(int j=0; size_t(j)<uOps2.size(); j++)
      {
      for(int k=0; size_t(k)<oOps.size(); k++)
        {
        if((uOps1[i] == uOps2[j]) && (uOps1[i] == oOps[k]))
          {                          //For those that are common:
          Ograph newg1 = g1;         //make a new graph
          newg1.revAll(uOps1[i]);    //that we do the unsorting inversion on.
          Ograph newg2 = g2;         //make a new graph
          newg2.revAll(uOps2[j]);    //that we do the unsorting inversion on.
          Ograph newidg = idg;       //make a new graph
          newidg.revAll(oOps[k]);    //that we do the unsorting inversion on.

                                                        //There's no escape path
          if(!printSigLattice(newg1, newg2, newidg, depth))
            {
            printDeadend(newg1, newg2, newidg, depth+1);//so continue down this.
  
            char *g1str = g1.getSeqStr();
            char *g2str = g2.getSeqStr();
            char *idgstr = idg.getSeqStr();
            char *ng1str = newg1.getSeqStr();
            char *ng2str = newg2.getSeqStr();
            char *nidgstr = newidg.getSeqStr();
            outputLink(ng1str, ng2str, nidgstr, newidg.getDistNH(),
                       g1str, g2str, idgstr, idg.getDistNH(), depth);
            free(g1str); free(g2str); free(ng1str); free(ng2str);
            free(idgstr); free(nidgstr);
            }
          }
        }
      }
    }
  }



/*Print the output graphviz (dot) link.
*/
void outputLink(const char* g1str, const char* g2str, const char* idgstr,
                int iddist, const char* ng1str, const char* ng2str,
                const char* nidgstr, int niddist, const int depth)
  {
  if(PRINT_ALL)
    {
    char pstr[strlen(ng1str) + strlen(ng2str) + strlen(nidgstr)+20];
    sprintf(pstr, "\"%s\\n%s\\n%s\\n(%d)\"", ng1str, ng2str, nidgstr, niddist); 

    if(depth == 0)
      cout << "  " << pstr << "[peripheries=3]" << endl;
    cout << "  " << pstr
         << " -> "
         << "\"" << g1str << "\\n" << g2str << "\\n" << idgstr << "\\n("
         << iddist << ")\""
         << ";\n";
    }
  else
    {
    char pstr[strlen(nidgstr)+20];
    sprintf(pstr, "\"%s\\n(%d)\"", nidgstr, niddist); 

    if(depth == 0)
      cout << "  " << pstr << "[peripheries=3]" << endl;
    cout << "  " << pstr
         << " -> "
         << "\"" << idgstr << "\\n(" << iddist << ")\""
         << ";\n";
    }
  }


/*Return the vector of common pairs (inversions) to g1 and g2 that yeild the
most common pairs in the resulting permutations.  Return the in-degree of
the ops returned in -ret-.
  ** Could be made more efficient with sorting.
*/
Pair getSingleBestCommon(Ograph &g1, Ograph &g2)
  {
  vector<Pair> oOps1, oOps2;
  Pair ret;
  int bestcount=0;

  g1.getSafeInvs(oOps1);       //Get the operations on divergent edges.
  g2.getSafeInvs(oOps2);       //Get the operations on divergent edges.

  for(int i=0; size_t(i)<oOps1.size(); i++)
    {                              //for each sorting op
    for(int j=0; size_t(j)<oOps2.size(); j++)
      {
      if(oOps1[i] == oOps2[j])     //that matches one in s2
        {
        Ograph newg1 = g1;         //make a new graph
        newg1.revAll(oOps1[i]);    //that we do the inversion on.
        Ograph newg2 = g2;         //make a new graph
        newg2.revAll(oOps2[j]);    //that we do the inversion on.

        vector<Pair> noOps1, noOps2;
        newg1.getSafeInvs(noOps1);
        newg2.getSafeInvs(noOps2);
        int count=0;               //Count common ops between the new ones:
        for(int k=0; size_t(k)<noOps1.size(); k++)
          for(int l=0; size_t(l)<noOps2.size(); l++)
            if(noOps1[k] == noOps2[l])
              count++;

        if(count > bestcount)      //If we found more shared sorting inversions
          {
          ret = oOps1[i];          //then record the inversion and
          bestcount = count;       //update the bestcount before.
          }
        }
      }
    }
  if(bestcount == 0)
    return Pair(0,0);

  return ret;
  }


/*Return the vector of common pairs (inversions) to g1 and g2 that yeild the
most common pairs in the resulting permutations.  Return the in-degree of
the ops returned in -ret-.
  ** Could be made more efficient with sorting.
*/
int getBestCommon(Ograph &g1, Ograph &g2, vector<Pair> &ret)
  {
  vector<Pair> oOps1, oOps2;
  int bestcount=0;

  g1.getSafeInvs(oOps1);       //Get the operations on divergent edges.
  g2.getSafeInvs(oOps2);       //Get the operations on divergent edges.

  for(int i=0; size_t(i)<oOps1.size(); i++)
    {                              //for each sorting op
    for(int j=0; size_t(j)<oOps2.size(); j++)
      {
      if(oOps1[i] == oOps2[j])     //that matches one in s2
        {
        Ograph newg1 = g1;         //make a new graph
        newg1.revAll(oOps1[i]);    //that we do the inversion on.
        Ograph newg2 = g2;         //make a new graph
        newg2.revAll(oOps2[j]);    //that we do the inversion on.

        vector<Pair> noOps1, noOps2;
        newg1.getSafeInvs(noOps1);
        newg2.getSafeInvs(noOps2);
        int count=0;               //Count common ops between the new ones:
        for(int k=0; size_t(k)<noOps1.size(); k++)
          for(int l=0; size_t(l)<noOps2.size(); l++)
            if(noOps1[k] == noOps2[l])
              count++;

        if(count > bestcount)      //If we found more shared sorting inversions
          {
          ret.clear();             //drop the others,
          bestcount = count;       //update the bestcount before...
          }
        if(count >= bestcount)     //if we found at least as many
          ret.push_back(oOps1[i]); //then record the inversion.
        }
      }
    }
  return bestcount;
  }



/*Return the vector of common pairs(inversions) for all in -graphs- that yeild
the most common pairs in the resulting permutations.  Return the in-degree of
the ops returned in -ret-.
*/
int getBestCommon(vector<Ograph> &graphs, vector<Pair> &ret)
  {
  vector<Pair> inter;
  getOpIntersection(inter, graphs);

  size_t bestcount=0;
  for(int i=0; size_t(i)<inter.size(); i++)
    {                             //For each sorting op in the intersection
    vector<Pair> inter2;          //Get the intersection with the op appied:
    for(int j=0; size_t(j)<graphs.size(); j++)
      graphs[j].revAll(inter[i]);     //Apply the inversion to all the graphs,
    getOpIntersection(inter2, graphs); //get the intersection,
    for(int j=0; size_t(j)<graphs.size(); j++)
      graphs[j].revAll(inter[i]);     //and un apply the inversion.

    if(inter2.size() > bestcount) //If we found more shared sorting inversions
      {
      ret.clear();                //drop the others,
      bestcount = inter2.size();  //update the bestcount before...
      }
    if(inter2.size() >= bestcount)//if we found at least as many
      ret.push_back(inter[i]);    //then record the inversion.
    }

  return bestcount;
  }


/*Return a vector that is the intesection of all the sorting ops for the
given vector of graphs.
*/
void getOpIntersection(vector<Pair> &inter, const vector<Ograph> &graphs)
  {
  vector<Pair> oOps;
                    //Find the intersection of sorting ops over all graphs:
  graphs[0].getSafeInvs(inter);          //Get the ops on divergent edges.
  for(int i=1; size_t(i)<graphs.size(); i++)
    {
    graphs[i].getSafeInvs(oOps);         //Get the ops on divergent edges.
    intersect(inter, oOps);
    oOps.clear();
    }
  }


/* Make the first vector the instersection of the two given vectors.
*/
void intersect(vector<Pair> &inter, const vector<Pair> &v)
  {
  vector<Pair>::iterator i = inter.begin();

  while(i != inter.end())
    {
    if(!isin(*i, v))
      i = inter.erase(i);
    else
      i++;
    }
  }


/* Return true if the given Pair is in the given vector.
*/
bool isin(const Pair &inter, const vector<Pair> &v)
  {
  for(int i=0; size_t(i) < v.size(); i++)
    if(v[i] == inter)
      return true;

  return false;
  }
