/* Krister Swenson                                                  Spring 2006

Function used for finding signatures.
 ** NEED TO MODIFY:  the multiple signature function take the inverse of the
                     graphs needed.
*/
#ifndef SIGFUNCS
#define SIGFUNCS

#include <iostream>
#include <vector>
#include <ext/hash_set>
#include <ext/hash_map>
#include "../graph/Pair.h"
#include "../graph/med.h"
#include "../seqtools/seqfuncs.h"

#define PRINT_ALL 0  //Print all three sequences if this is 1.


  //Return the size of the largest signature between -g1- and -g2-.
int getSigSize(Ograph &g1, Ograph &g2, const Ograph &idg);
int getSigSize_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth);

  //Return a maximum signature between -g1- and -g2-.
Ograph getSig(Ograph &g1, Ograph &g2, const Ograph &idg);
Ograph getSig_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth);
  //Return all maximum signature between -g1- and -g2-.
vector<Ograph> getSigs(Ograph &g1, Ograph &g2, const Ograph &idg);
vector<Ograph> getSigs_R(Ograph &g1, Ograph &g2, const Ograph &idg,
                         const int depth);

  //Return the depth(size) of the largest signature found via the greedy method.
int getGreedySig(Ograph &g1, Ograph &g2, const Ograph &idg);
int getGreedySig_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth);

  //Return the depth(size) of the largest signature found via the greedy method.
  // This version is faster because it keeps track of the sequences already
  // visited and truncates the upwards push if we've already been there.
int getGreedySigF(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth=0);
int getGreedySigF_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth);

  //Return the depth of the largest signature found via the greedy method where
  // the first tie is always chosen.
int getGreedySigSP(Ograph &g1, Ograph &g2, const Ograph &idg);
int getGreedySigSP_R(Ograph &g1, Ograph &g2, const Ograph &idg, const int depth);

  //Return the depth of the largest signature found for the given set of 
  // sequences via the greedy method where the first tie is always chosen.
int getGreedySigSP(std::vector<Ograph> &graphs);
int getGreedySigSP(std::vector<Ograph> &graphs, const Ograph &idg,
                   const int depth=0);



  //Print every node of the lattice that is a signature.
  //For searching for paths of -depth- up the lattice the value is the
  //depth to truncate the upwards push at; if negative we just print the
  //lattice.  Immediately stop search and return true if the upwards push was
  //truncated.
bool printSigLattice(Ograph &g1,Ograph &g2,const Ograph &idg,const int depth);

  //Print every node of the signature path that maximizes the number of shared
  //common sorting inversion in the resulting permutation.
  //-depth- should always be given as 0.
  //Return the depth of the largest signature found.
int printGreedy(Ograph &g1, Ograph &g2, const Ograph &idg, int depth);

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


  //Return the vector of common pairs (inversions) to g1 and g2 that yeild the
  //most common pairs in the resulting permutations.  Return the in-degree of
  //the ops returned in -ret-.
int getBestCommon(Ograph &g1, Ograph &g2, std::vector<Pair> &ret);
  //Return the vector of common pairs(inversions) for all in -graphs- that yeild
  //the most common pairs in the resulting permutations.  Return the in-degree
  //of the ops returned in -ret-.
int getBestCommon(std::vector<Ograph> &graphs, std::vector<Pair> &ret);

  //Return a vector that is the intesection of all the sorting ops for the
  //given vector of graphs.
void getOpIntersection(std::vector<Pair> &inter,
                       const std::vector<Ograph> &graphs);


  //Return a common pair (of inversions) to g1 and g2 that yeild the
  //most common pairs in the resulting permutations.
Pair getSingleBestCommon(Ograph &g1, Ograph &g2);


  //Print the output graphviz (dot) link.
void outputLink(const char* g1str, const char* g2str, const char* idgstr,
                int iddist, const char* ng1str, const char* ng2str,
                const char* nidgstr, int niddist, const int depth);

  //Print the dead-end path down as far as possible towards the inverse
  //identity.  In other word, unsort the sequences so that they don't gain some
  //path that can move them upwards through the lattice and past the starting
  //spot.  The -depth- should be the upwards distance past the given sequence
  //the upwards search should stop at, normally 1.
void printDeadend(Ograph &g1, Ograph &g2, Ograph &idg, const int depth);

  //Make the first vector the instersection of the two given vectors.
void intersect(std::vector<Pair> &inter, const std::vector<Pair> &v);

  //Return true if the given Pair is in the given vector.
bool isin(const Pair &inter, const std::vector<Pair> &v);




  //Comparison function for a hash_map on a string.
struct eqstr
  {
  bool operator()(const char* s1, const char* s2) const
    { return strcmp(s1, s2) == 0; }
  };


  //Lookup function for the beenseen hash_set.
bool lookup(const __gnu_cxx::hash_set
                         <const char*,__gnu_cxx::hash<const char*>, eqstr>& Set,
            const char* word);


#endif
