/* Krister Swenson                                                 Winter 2010
 *
 * This is a trace kept in normal form.
*/
#ifndef TRACE_H
#define TRACE_H

#include <iostream>
#include <set>
#include <vector>
#include <list>
#include <string>
#include "Word.h"



/*___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___
Type definitions:                                                             */




/*___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___
Trace definition:                                                             */


class Trace
  {
  friend ostream& operator<<(ostream& output, const Trace& t);

  public:
    Trace();
      //Create a trace with one word.
    Trace(const Word &w);
      //Create a trace from the given word (list of sets).
    Trace(const std::list<Element> &invs);
      //Create a normal form by appending the given word to the given trace.
    Trace(const Trace &t, const Word &w);
    Trace(const Trace &t);
    void copy(const Trace &t);
    void operator=(const Trace &t);


      //Append the given word to the Trace.
    void appendWord(const Word &w);
      //Add the given element to the normal form.
    void addElement(const Element &e);
      //Return true if the given element would be the last element of the
      //trace if it were added.
    bool goesAtEnd(const Element &e) const;
      //Return true if the given element belongs in the first word.
    bool goesInFirstWord(const Element &e) const;

      //Modify the normal form as though it were sorting from the identity to
      // the given permutation.
    void invert(const std::vector<int> &p);
      //Reverse the order of the normal form.
    void reverse();
      //Rename the normal so that the sequence corresponds to sorting from the
      // identity to the given permutation.
    void rename(const std::vector<int> &p);


      //Remove the last element from the Trace.
    void removeLastElement();
      //Get the last element from the Trace.
    Element getLastElement() const;
      //Remove the last word from the Trace.
    void removeLast();
      //Remove all but the first word.
    void removeAllButFirst();

    string getString() const;

  private:

    list<Word> normalform;   //The trace.
  };


#endif
